package org.jythonx.modules;

import org.python.core.*;
import org.jythonx.utils.*;

public class _heapqmodule{

    static int _siftdown(PyList heap, int startpos, int pos){
	PyObject newitem, parent;
	int cmp;
	int parentpos;

	if (pos >= heap.__len__()){
            throw Py.IndexError("index out of range");
	}

	newitem = heap.__getitem__(pos);
	/* Follow the path to the root, moving parents down until finding
	   a place newitem fits. */
	while (pos > startpos){
	    parentpos = (pos - 1) >> 1;
	    parent = heap.__getitem__(parentpos);
	    cmp = PyFunctions.PyObject_RichCompareBool(parent, newitem, PyFunctions.LE);
	    if (cmp == -1) {
                PyFunctions.ThrowErrorIfSet();
		return -1;
	    }
	    if (cmp == 1)
	        break;
            heap.__setitem__(pos, parent);
	    pos = parentpos;
	}
        heap.__setitem__(pos, newitem);
	return 0;
    }



    static int _siftup(PyList heap, int pos){
	int startpos, endpos, childpos, rightpos;
	int cmp;
	PyObject newitem, tmp;

	endpos = heap.__len__();
	startpos = pos;
	if (pos >= endpos) {
            throw Py.IndexError("index out of range");
	}
	newitem = heap.__getitem__(pos);

	/* Bubble up the smaller child until hitting a leaf. */
	childpos = 2*pos + 1;    /* leftmost child position  */
	while (childpos < endpos) {
	    /* Set childpos to index of smaller child.   */
	    rightpos = childpos + 1;
	    if (rightpos < endpos) {
	        cmp = PyFunctions.PyObject_RichCompareBool(heap.__getitem__(rightpos),
							   heap.__getitem__(childpos),
                                                           PyFunctions.LE);
		if (cmp == -1) {
		    PyFunctions.ThrowErrorIfSet();
		    return -1;
		}
		if (cmp == 1)
		    childpos = rightpos;
	    }
	    /* Move the smaller child up. */
	    tmp = heap.__getitem__(childpos);
            heap.__setitem__(pos, tmp);
	    pos = childpos;
	    childpos = 2*pos + 1;
	}

	/* The leaf at pos is empty now.  Put newitem there, and and bubble
	   it up to its final resting place (by sifting its parents down). */
        heap.__setitem__(pos, newitem);
	return _siftdown(heap, startpos, pos);
    }

    public static PyObject heappush(PyObject[] args){
	PyObject heap = null, item = null;

        if(args.length != 2)
            throw Py.TypeError("heappush expected 2 arguments, got " + args.length);
        heap = args[0];
        item = args[1];
	if (!(heap instanceof PyList)) {
	    throw Py.TypeError("heap argument must be a list");
	}

        PyList heap_ = (PyList)heap;
        heap_.append(item);
	if (_siftdown(heap_, 0, (heap_.__len__()-1)) == -1)
	    return null;
	return Py.None;
    }



    public static PyObject heappop(PyObject heap){
	PyObject lastelt, returnitem;
	int n;

	if (!(heap instanceof PyList)) {
	    throw Py.TypeError("heap argument must be a list");
	}

	/* # raises appropriate IndexError if heap is empty */
	n = heap.__len__();
	if (n == 0) {
	    throw Py.IndexError("index out of range");
	}

	lastelt = heap.__getitem__(n - 1);// PyList_GET_ITEM(heap, n-1) ;
        heap.__delslice__(new PyInteger(n -1), new PyInteger(n));
	n--;

        if(n == 0)
            return lastelt;
	returnitem = heap.__getitem__(0);
        heap.__setitem__(0, lastelt);
	if (_siftup((PyList)heap, 0) == -1) {
	    return null;
	}
	return returnitem;
    }

    public static PyObject heapreplace(PyObject[] args){
	PyObject heap, item, returnitem;

        if(args.length != 2)
            throw Py.TypeError("heapreplace expected 2 arguments, got " + args.length);
        heap = args[0];
        item = args[1];

        if(!(heap instanceof PyList)){
	    throw Py.TypeError("heap argument must be a list");
	}

	if (heap.__len__() < 1) {
	    throw Py.IndexError("index out of range");
	}

	returnitem = heap.__getitem__(0);
        heap.__setitem__(0, item);
	if (_siftup((PyList)heap, 0) == -1) {
	    return null;
	}
	return returnitem;
    }

    public static PyObject heapify(PyObject heap){
	int i, n;

        if(!(heap instanceof PyList)){
	    throw Py.TypeError("heap argument must be a list");
	}

	n = heap.__len__();
	/* Transform bottom-up.  The largest index there's any point to
	   looking at is the largest with a child index in-range, so must
	   have 2*i + 1 < n, or i < (n-1)/2.  If n is even = 2*j, this is
	   (2*j-1)/2 = j-1/2 so j-1 is the largest, which is n//2 - 1.  If
	   n is odd = 2*j+1, this is (2*j+1-1)/2 = j so j-1 is the largest,
	   and that's again n//2-1.
	*/
        PyList heap_ = (PyList)heap;
	for (i=n/2-1 ; i>=0 ; i--)
	    if(_siftup(heap_, i) == -1)
	        return null;
	return Py.None;
    }

    public static PyObject nlargest(PyObject[] args){
	PyObject elem, iterable, sol, it, oldelem;
	int i, n;

        if(args.length != 2){
            throw Py.TypeError("nlargest expected 2 arguments, got " + args.length);
        }
        n = ((PyInteger)args[0].__int__()).getValue();
        iterable = args[1];

	it = iterable.__iter__();
	if (it == null)
	     return null;

	PyList heap = new PyList();//PyList_New(0);

        sortit:{
	    for (i=0 ; i<n ; i++ ){
	        elem = PyFunctions.PyIter_Next(it);
		if (elem == null) {
		    if (PyFunctions.PyErr_Occurred())
		        PyFunctions.ThrowErrorIfSet();
		    else
		       break sortit;
		}
                heap.append(elem);
	    }
            if(heap.__len__() == 0)
		break sortit;

	    for (i=n/2-1 ; i>=0 ; i--)
		if(_siftup(heap, i) == -1)
		    return null;

	    sol = heap.__getitem__(0);
	    while (true) {
		elem = PyFunctions.PyIter_Next(it);
		if (elem == null) {
			if (PyFunctions.PyErr_Occurred())
			    PyFunctions.ThrowErrorIfSet();
			else
			    break sortit;
		}
		if (PyFunctions.PyObject_RichCompareBool(elem, sol, PyFunctions.LE) != 0) {
                    PyFunctions.ThrowErrorIfSet();
		    continue;
		}
		oldelem = heap.__getitem__(0);
                heap.__setitem__(0, elem);
		if (_siftup(heap, 0) == -1)
		    return null;
		sol = heap.__getitem__(0);
	    }
        }//sortit:
        heap.sort();
        heap.reverse();
	return heap;
    }

    static int _siftdownmax(PyList heap, int startpos, int pos){
	PyObject newitem, parent;
	int cmp;
	int parentpos;

        if(pos >= heap.__len__()){
            throw Py.IndexError("index out of range");
	}

	newitem = heap.__getitem__(pos);// PyList_GET_ITEM(heap, pos);
	/* Follow the path to the root, moving parents down until finding
	   a place newitem fits. */
	while (pos > startpos){
		parentpos = (pos - 1) >> 1;
		parent = heap.__getitem__(parentpos);
		cmp = PyFunctions.PyObject_RichCompareBool(newitem, parent, PyFunctions.LE);
		if (cmp == -1) {
                    PyFunctions.ThrowErrorIfSet();
		    return -1;
		}
		if (cmp == 1)
		    break;
                heap.__setitem__(pos, parent);
		pos = parentpos;
	}
        heap.__setitem__(pos, newitem);
	return 0;
    }

    static int _siftupmax(PyList heap, int pos){
        int startpos, endpos, childpos, rightpos;
	int cmp;
	PyObject newitem, tmp;

	endpos = heap.__len__();
	startpos = pos;
	if (pos >= endpos) {
	    throw Py.IndexError("index out of range");
	}
	newitem =  heap.__getitem__(pos);

	/* Bubble up the smaller child until hitting a leaf. */
	childpos = 2*pos + 1;    /* leftmost child position  */
	while (childpos < endpos) {
		/* Set childpos to index of smaller child.   */
		rightpos = childpos + 1;
		if (rightpos < endpos) {
			cmp = PyFunctions.PyObject_RichCompareBool(
				heap.__getitem__(childpos),
                                heap.__getitem__(rightpos),
                                PyFunctions.LE);
			if (cmp == -1) {
                            PyFunctions.ThrowErrorIfSet();
			    return -1;
			}
			if (cmp == 1)
			    childpos = rightpos;
		}
		/* Move the smaller child up. */
		tmp = heap.__getitem__(childpos);
                heap.__setitem__(pos, tmp);
		pos = childpos;
		childpos = 2*pos + 1;
	}

	/* The leaf at pos is empty now.  Put newitem there, and and bubble
	   it up to its final resting place (by sifting its parents down). */
        heap.__setitem__(pos, newitem);
	return _siftdownmax(heap, startpos, pos);
    }



    public static PyObject nsmallest(PyObject[] args){
	PyObject elem, los, it, oldelem;
	int i;

        if(args.length != 2)
            throw Py.TypeError("nsmallest expected 2 arguments, got " + args.length);
        int n = ((PyInteger)args[0].__int__()).getValue();
	PyObject iterable = args[1];

        

	it = iterable.__iter__();
	if (it == null)
	    return null;
	PyList heap = new PyList();
        sortit:{
	    for (i=0 ; i<n ; i++ ){
	        elem = PyFunctions.PyIter_Next(it);
	        if (elem == null) {
	            if (PyFunctions.PyErr_Occurred())
		        PyFunctions.ThrowErrorIfSet();
		    else
		        break sortit;
	        }
	        heap.append(elem);
           }
           n = heap.__len__();
	   if (n == 0)
	       break sortit;

	   for (i=n/2-1 ; i>=0 ; i--)
	       if(_siftupmax(heap, i) == -1)
	          return null;

	   los = heap.__getitem__(0);
	   while (true) {
	       elem = PyFunctions.PyIter_Next(it);
	        if (elem == null) {
                    if(PyFunctions.PyErr_Occurred())
		        PyFunctions.ThrowErrorIfSet();
		    else
		        break sortit;
	        }
                if(PyFunctions.PyObject_RichCompareBool(los, elem, PyFunctions.LE) != 0){
                    PyFunctions.ThrowErrorIfSet();
		    continue;
	        }

	        oldelem = heap.__getitem__(0);
	        heap.__setitem__(0, elem);
	        if (_siftupmax(heap, 0) == -1)
	            return null;
	        los =  heap.__getitem__(0);
	    }
        }//sortit:
        heap.sort();
	return heap;
    }
}