package org.jythonx.bzip2;

import java.io.*;
import java.nio.*;
import java.nio.channels.*;

public class BZLib implements BZConstants{

    public static String BZ_VERSION = "1.0.4, 20-Dec-2006";
    public static long BZ_INITIALISE_CRC(){ return 0xffffffffL; }

    public static void prepare_new_block(EState s ){
         s.nblock = 0;
         s.numZ = 0;
         s.state_out_pos = 0;
         s.blockCRC = BZ_INITIALISE_CRC();
         for (int i = 0; i < 256; i++) s.inUse[i] = false;
         s.blockNo++;
    }

    public static void init_RL (EState s){
        s.state_in_ch  = 256;
        s.state_in_len = 0;
    }

    static boolean isempty_RL (EState s){
        if (s.state_in_ch < 256 && s.state_in_len > 0)
            return false; 
        else
            return true;
    }

    public static int BZ2_bzCompressInit( 
                     bz_stream strm, 
                     int        blockSize100k,
                     int        verbosity,
                     int        workFactor ){

        if (strm == null || 
            blockSize100k < 1 || blockSize100k > 9 ||
            workFactor < 0 || workFactor > 250)
                return BZ_PARAM_ERROR;

        if (workFactor == 0) workFactor = 30;
        EState s = new EState();
        s.strm = strm;

        int n       = 100000 * blockSize100k;
        s.arr1 = new IntView(new byte[n * 4]);
        s.arr2 = new IntView(new byte[(n+ BZ_N_OVERSHOOT) * 4]);
        s.ftab = new IntView(new byte[65537 * 4]);

        s.blockNo           = 0;
        s.state             = BZ_S_INPUT;
        s.mode              = BZ_M_RUNNING;
        s.combinedCRC       = 0;
        s.blockSize100k     = blockSize100k;
        s.nblockMAX         = 100000 * blockSize100k - 19;
        s.verbosity         = verbosity;
        s.workFactor        = workFactor;

        s.block             = s.arr2.b;//(UChar*)s->arr2;
        s.mtfv              = new ShortView(s.arr1.b);//s.arr1;// (UInt16*)s->arr1;
        s.zbits             = null;//NULL;
        s.ptr               = s.arr1;// (UInt32*)s->arr1;

        strm.state          = s;
        strm.total_in_lo32  = 0;
        strm.total_in_hi32  = 0;
        strm.total_out_lo32 = 0;
        strm.total_out_hi32 = 0;
        init_RL ( s );
        prepare_new_block ( s );
        return BZ_OK;
    }

    public static long BZ_UPDATE_CRC(long crcVar, byte cha){                          
	return  (((crcVar << 8) & UINT32) ^                    
		 BZ2_crc32Table[(int)(((crcVar >> 24) ^ cha) & 0xff)]) & UINT32;      
    }

    public static void add_pair_to_block (EState s){
        int i;
        byte ch = (byte)(s.state_in_ch & 0xff);
        for (i = 0; i < s.state_in_len; i++) {
            s.blockCRC = BZ_UPDATE_CRC(s.blockCRC, ch);
        }
        s.inUse[(int)(s.state_in_ch & 0xff)] = true;
        switch (s.state_in_len) {
            case 1:
                s.block[s.nblock] = ch; s.nblock++;
                break;
            case 2:
                s.block[s.nblock] = ch; s.nblock++;
                s.block[s.nblock] = ch; s.nblock++;
                break;
            case 3:
                s.block[s.nblock] = ch; s.nblock++;
                s.block[s.nblock] = ch; s.nblock++;
                s.block[s.nblock] = ch; s.nblock++;
                break;
            default:
                s.inUse[s.state_in_len-4] = true;
                s.block[s.nblock] = ch; s.nblock++;
                s.block[s.nblock] = ch; s.nblock++;
                s.block[s.nblock] = ch; s.nblock++;
	        s.block[s.nblock] = ch; s.nblock++;
                s.block[s.nblock] = ((byte)(s.state_in_len-4));
                s.nblock++;
                break;
        }
    }

    public static void flush_RL (EState s ){
        if (s.state_in_ch < 256) add_pair_to_block (s);
        init_RL (s);
    }

    public static void ADD_CHAR_TO_BLOCK(EState zs, byte zchh0){
        int zchh = (int)zchh0;
        /*-- fast track the common case --*/           
        if (zchh != zs.state_in_ch && zs.state_in_len == 1) {                 
            byte ch = (byte)((zs.state_in_ch & 0xff));
            zs.blockCRC = BZ_UPDATE_CRC(zs.blockCRC, ch) & UINT32; 
            zs.inUse[(int)(zs.state_in_ch & 0xff)] = true;  
            zs.block[zs.nblock] = ch;
            zs.nblock++;                               
            zs.state_in_ch = zchh;                     
        }                                  
        else{
            /*-- general, uncommon cases --*/              
            if (zchh != zs.state_in_ch || zs.state_in_len == 255) {
	        if ((zs.state_in_ch) < 256)                  
                    add_pair_to_block (zs); 
                zs.state_in_ch = zchh;                     
                zs.state_in_len = 1;                       
            } else {                                       
                zs.state_in_len++;                         
            } 
	 }  
    }

    public static boolean copy_input_until_stop (EState s){
        boolean progress_in = false;
        bz_stream strm = s.strm;
        byte[] next_in = strm._next_in;
        if (s.mode == BZ_M_RUNNING) {
            /*-- fast track the common case --*/
            while (true) {
                /*-- block full? --*/
                if (s.nblock >= s.nblockMAX) break;
                /*-- no input? --*/
                if (strm.avail_in == 0){
                    break;
                }
                progress_in = true;
                ADD_CHAR_TO_BLOCK(s, next_in[strm._ni++]);
                //s.strm.inc_next_in();
                strm.avail_in--;
                strm.total_in_lo32++;
                if (strm.total_in_lo32 == 0) 
                    strm.total_in_hi32++;
           }
        } else {
            /*-- general, uncommon case --*/
            while (true) {
                /*-- block full? --*/
                if (s.nblock >= s.nblockMAX) break;
                /*-- no input? --*/
                if (s.strm.avail_in == 0) break;
                /*-- flush/finish end? --*/
                if (s.avail_in_expect == 0) break;
                progress_in = true;
                ADD_CHAR_TO_BLOCK(s, next_in[strm._ni++]);
                //s.strm.inc_next_in();
                strm.avail_in--;
                strm.total_in_lo32++;
                if (strm.total_in_lo32 == 0) strm.total_in_hi32++;
                s.avail_in_expect--;
            }
        }
        return progress_in;
    }

    public static boolean copy_output_until_stop(EState s){
        boolean progress_out = false;
        while (true) {
            /*-- no output space? --*/
            if (s.strm.avail_out == 0){
                 break;
            }

            /*-- block done? --*/
	    if (s.state_out_pos >= s.numZ){
                break;
            }

            progress_out = true;
            s.strm.put_next_out(s.zbits[s.state_out_pos + s.zbstart]);
            s.state_out_pos++;
            s.strm.avail_out--;
            s.strm.inc_next_out();
            s.strm.total_out_lo32++;
            if (s.strm.total_out_lo32 == 0) s.strm.total_out_hi32++;
        }
        return progress_out;
    }

    public static boolean handle_compress(bz_stream strm){
        boolean progress_in  = false;
        boolean progress_out = false;
        EState s = (EState)strm.state;   
        int hcloop = 0;
        while (true) {
            if (s.state == BZ_S_OUTPUT) {
                progress_out |= copy_output_until_stop ( s );
	        if (s.state_out_pos < s.numZ){
                    break;
	        }
                if (s.mode == BZ_M_FINISHING && 
                    s.avail_in_expect == 0 &&
		    isempty_RL(s)){ 
                     break;
                }
                prepare_new_block ( s );
                s.state = BZ_S_INPUT;
                if (s.mode == BZ_M_FLUSHING && 
                    s.avail_in_expect == 0 &&
                    isempty_RL(s)) break; 
	    }

            if (s.state == BZ_S_INPUT) {
                progress_in |= copy_input_until_stop ( s );
                if (s.mode != BZ_M_RUNNING && s.avail_in_expect == 0) {
                    flush_RL ( s );
                    compress.BZ2_compressBlock ( s, s.mode == BZ_M_FINISHING );
                   s.state = BZ_S_OUTPUT;
                }
                else{
                    if (s.nblock >= s.nblockMAX) {
                        compress.BZ2_compressBlock ( s, false );
                        s.state = BZ_S_OUTPUT;
                    }
                    else{
                        if (s.strm.avail_in == 0) {
                            break;
                        }
		    } 
	        }
	    }
       }
       return progress_in || progress_out;
    }


    public static int BZ2_bzCompress(bz_stream strm, int action){
        boolean progress;
        if (strm == null) return BZ_PARAM_ERROR;
        EState s = (EState)strm.state;
        if (s == null) return BZ_PARAM_ERROR;
        if (s.strm != strm) return BZ_PARAM_ERROR;
        boolean preswitch = false;
        preswitch:
        do{
	    preswitch = false;
            switch (s.mode) {
                case BZ_M_IDLE:
                    return BZ_SEQUENCE_ERROR;

                case BZ_M_RUNNING:
                   if (action == BZ_RUN) {
                       progress = handle_compress ( strm );
                       return progress ? BZ_RUN_OK : BZ_PARAM_ERROR;
                   } 
                   else
	               if (action == BZ_FLUSH) {
                          s.avail_in_expect = strm.avail_in;
                          s.mode = BZ_M_FLUSHING;
                          preswitch = true;
                          continue preswitch;
                       }
                       else
                           if (action == BZ_FINISH) {
                               s.avail_in_expect = strm.avail_in;
                               s.mode = BZ_M_FINISHING;
                               preswitch = true;
                               continue preswitch;
                           }
                           else 
                               return BZ_PARAM_ERROR;
                case BZ_M_FLUSHING:
                    if (action != BZ_FLUSH) return BZ_SEQUENCE_ERROR;
                    if (s.avail_in_expect != s.strm.avail_in) 
                        return BZ_SEQUENCE_ERROR;
                    progress = handle_compress ( strm );
                    if (s.avail_in_expect > 0 || !isempty_RL(s) ||
                        s.state_out_pos < s.numZ) return BZ_FLUSH_OK;
                    s.mode = BZ_M_RUNNING;
                    return BZ_RUN_OK;
                case BZ_M_FINISHING:
                    if (action != BZ_FINISH) return BZ_SEQUENCE_ERROR;
                    if (s.avail_in_expect != s.strm.avail_in) 
                        return BZ_SEQUENCE_ERROR;
                    progress = handle_compress ( strm );
                    if (!progress) return BZ_SEQUENCE_ERROR;
                    if (s.avail_in_expect > 0 || !isempty_RL(s) ||
                        s.state_out_pos < s.numZ) return BZ_FINISH_OK;
                    s.mode = BZ_M_IDLE;
                    return BZ_STREAM_END;
             }
         } while(preswitch);
         return BZ_OK; /*--not reached--*/
    }

    public static int BZ2_bzCompressEnd(bz_stream strm){
        if (strm == null) return BZ_PARAM_ERROR;
        EState s = (EState)strm.state;
        if (s == null) return BZ_PARAM_ERROR;
        if (s.strm != strm) return BZ_PARAM_ERROR;

        if (s.arr1 != null) s.arr1 = null;
        if (s.arr2 != null) s.arr2 = null;
        if (s.ftab != null) s.ftab = null;
        strm.state = null;
        //strm->state = NULL;   
        return BZ_OK;
    }

    public static bzFile BZ2_bzWriteOpen 
                    ( int[]  bzerror,      
                      WritableByteChannel f, 
                      int   blockSize100k, 
                      int   verbosity,
                      int   workFactor ){
        int   ret;
        bzFile bzf = null;

        BZ_SETERR(bzf, BZ_OK, bzerror);

        if (f == null ||
            (blockSize100k < 1 || blockSize100k > 9) ||
            (workFactor < 0 || workFactor > 250) ||
            (verbosity < 0 || verbosity > 4))
	    { BZ_SETERR(bzf, BZ_PARAM_ERROR, bzerror); return null; };

        //if (ferror(f))
        // { BZ_SETERR(BZ_IO_ERROR); return NULL; };

        bzf = new bzFile();//malloc ( sizeof(bzFile) );
        //if (bzf == NULL)
        // { BZ_SETERR(BZ_MEM_ERROR); return NULL; };

        BZ_SETERR(bzf, BZ_OK, bzerror);
        bzf.initialisedOk = false;
        bzf.bufN          = 0;
        bzf.handle        = f;
        bzf.writing       = true;
        bzf.strm.bzalloc  = null;
        bzf.strm.bzfree   = null;
        bzf.strm.opaque   = null;

        if (workFactor == 0) workFactor = 30;
        ret = BZ2_bzCompressInit ( bzf.strm, blockSize100k, 
                                  verbosity, workFactor );
        if (ret != BZ_OK)
        { 
	    BZ_SETERR(bzf, ret, bzerror); 
            //free(bzf); 
            return null; 
        };

        bzf.strm.avail_in = 0;
        bzf.initialisedOk = true;
        return bzf;   
    }

    public static void BZ2_bzWrite
                 ( int[]    bzerror, 
                   bzFile b, 
                   ByteBuffer   buf, 
                   int     len ) throws IOException{
        int n, n2, ret;
        bzFile bzf = b;

        BZ_SETERR(bzf, BZ_OK, bzerror);
        if (bzf == null || buf == null || len < 0)
	{ BZ_SETERR(bzf, BZ_PARAM_ERROR, bzerror); return; };
        if (!(bzf.writing))
	{ BZ_SETERR(bzf, BZ_SEQUENCE_ERROR, bzerror); return; };
        //if (ferror(bzf->handle))
        // { BZ_SETERR(BZ_IO_ERROR); return; };

        if (len == 0)
        { BZ_SETERR(bzf, BZ_OK, bzerror); return; };

        bzf.strm.avail_in = len;
        bzf.strm.next_in  = buf;
        bzf.strm._next_in = buf.array();
        bzf.strm._ni = 0;
        bzf.strm.next_in.position(0);

        WritableByteChannel handle = (WritableByteChannel)bzf.handle;
        while (true) {
            bzf.strm.avail_out = BZ_MAX_UNUSED;
            bzf.strm.next_out = bzf.buf;
            bzf.buf.clear();
            ret = BZ2_bzCompress ( bzf.strm, BZ_RUN );
            if (ret != BZ_RUN_OK)
	    { BZ_SETERR(bzf, ret, bzerror); return; };
            if (bzf.strm.avail_out < BZ_MAX_UNUSED) {
	        n = (int)((long)BZ_MAX_UNUSED - bzf.strm.avail_out);
                //n2 = fwrite ( (void*)(bzf->buf), sizeof(UChar), 
	        //            n, bzf->handle );
	        bzf.buf.flip();     
                n2 = handle.write(bzf.buf);
                bzf.buf.clear();
                if (n != n2)
	        { BZ_SETERR(bzf, BZ_IO_ERROR, bzerror); return; };
            }
            if (bzf.strm.avail_in == 0)
	    { BZ_SETERR(bzf, BZ_OK, bzerror); return; };
        }
    }

    public static void BZ2_bzWriteClose64
                   ( int[]          bzerror, 
                     bzFile       b, 
                     int           abandon, long[] nbytes) throws IOException {
        int   n, n2, ret;
        bzFile bzf = b;

        if (bzf == null)
	{ BZ_SETERR(bzf, BZ_OK, bzerror); return; };
        if (!(bzf.writing))
	{ BZ_SETERR(bzf, BZ_SEQUENCE_ERROR, bzerror); return; };
        //if (ferror(bzf->handle))
        // { BZ_SETERR(BZ_IO_ERROR); return; };

        //nbytes_in_lo32 = 0;
        nbytes[0] = 0;
        //nbytes_in_hi32 = 0;
        nbytes[1] = 0;
        //nbytes_out_lo32 = 0;
        nbytes[2] = 0;
        //nbytes_out_hi32 = 0;
        nbytes[3] = 0;

        if ((abandon==0) && bzf.lastErr == BZ_OK) {
            WritableByteChannel handle = (WritableByteChannel)bzf.handle;
            while (true) {
                bzf.strm.avail_out = BZ_MAX_UNUSED;
                bzf.strm.next_out = bzf.buf;
                bzf.buf.clear();
                ret = BZ2_bzCompress (bzf.strm, BZ_FINISH );
                if (ret != BZ_FINISH_OK && ret != BZ_STREAM_END)
		{ BZ_SETERR(bzf, ret, bzerror); return; };

                if (bzf.strm.avail_out < BZ_MAX_UNUSED) {
	            n = (int)(BZ_MAX_UNUSED - bzf.strm.avail_out);
                    //n2 = fwrite ( (void*)(bzf->buf), sizeof(UChar), 
	            //            n, bzf->handle );
                    bzf.buf.flip();
                    n2 = handle.write(bzf.buf);
                    bzf.buf.clear();
                    if (n != n2)
	            { BZ_SETERR(bzf, BZ_IO_ERROR, bzerror); return; }
                }
                if (ret == BZ_STREAM_END) break;
            }
        }

        //if ( !abandon && !ferror ( bzf->handle ) ) {
        // fflush ( bzf->handle );
        // if (ferror(bzf->handle))
        //    { BZ_SETERR(BZ_IO_ERROR); return; };
        //}

        //if (nbytes_in_lo32 != NULL)
        nbytes[0] = bzf.strm.total_in_lo32;
        //if (nbytes_in_hi32 != NULL)
        nbytes[1] = bzf.strm.total_in_hi32;
        //if (nbytes_out_lo32 != NULL)
        nbytes[2] = bzf.strm.total_out_lo32;
        //if (nbytes_out_hi32 != NULL)
        nbytes[3] = bzf.strm.total_out_hi32;

        BZ_SETERR(bzf, BZ_OK, bzerror);
        BZ2_bzCompressEnd (bzf.strm);
        //free ( bzf );
    }

    //DECOMPRESS...
    public static int BZ2_indexIntoF (int indx, int[] cftab ){
        int nb, na, mid;
        nb = 0;
        na = 256;
        do {
           mid = (nb + na) >> 1;
           if (indx >= cftab[mid]) nb = mid; else na = mid;
        }
        while (na - nb != 1);
       return nb;
    }

    public static int BZ2_bzDecompressInit
                      (bz_stream strm, 
                       int        verbosity,
                       boolean        small ){
        DState s;

        //if (!bz_config_ok()) return BZ_CONFIG_ERROR;

        if (strm == null) return BZ_PARAM_ERROR;
        //if (small != 0 && small != 1) return BZ_PARAM_ERROR;
        if (verbosity < 0 || verbosity > 4) return BZ_PARAM_ERROR;

        //if (strm.bzalloc == null) strm.bzalloc = default_bzalloc;
        //if (strm.bzfree == null) strm.bzfree = default_bzfree;

        s = new DState(); //BZALLOC( sizeof(DState) );
        s.strm                  = strm;
        strm.state              = s;
        s.state                 = BZ_X_MAGIC_1;
        s.bsLive                = 0;
        s.bsBuff                = 0;
        s.calculatedCombinedCRC = 0;
        strm.total_in_lo32      = 0;
        strm.total_in_hi32      = 0;
        strm.total_out_lo32     = 0;
        strm.total_out_hi32     = 0;
        s.smallDecompress       =  small;
        s.ll4                   = null;
        s.ll16                  = null;
        s.tt                    = null;
        s.currBlockNo           = 0;
        s.verbosity             = verbosity;

        return BZ_OK;
    }

    public static boolean unRLE_obuf_to_output_FAST (DState s){
        //UChar k1;
        byte k1;

        if (s.blockRandomised) {
            while (true) {
                /* try to finish existing run */
                while (true) {
                    if (s.strm.avail_out == 0) return false;
                    if (s.state_out_len == 0) break;
                    //s.strm.next_out = s.state_out_ch;
                    s.strm.put_next_out(s.state_out_ch);
                    s.calculatedBlockCRC = BZ_UPDATE_CRC ( s.calculatedBlockCRC, s.state_out_ch );
                    s.state_out_len--;
                    //s.strm.next_out++;
                    s.strm.inc_next_out();
                    s.strm.avail_out--;
                    s.strm.total_out_lo32++;
                    if (s.strm.total_out_lo32 == 0) s.strm.total_out_hi32++;
                }

                /* can a new run be started? */
                if (s.nblock_used == s.save_nblock+1) return false;
               
                /* Only caused by corrupt data stream? */
                if (s.nblock_used > s.save_nblock+1)
                    return true;
   
                s.state_out_len = 1;
                s.state_out_ch = (byte)s.k0;
                //BZ_GET_FAST(k1); 
                //s->tPos = s->tt[s->tPos];      \
                //cccc = (UChar)(s->tPos & 0xff);           \
                //s->tPos >>= 8;
                s.tPos = s.tt.getUInt32((int)s.tPos);//[(int)s.tPos];
                k1 = (byte)(s.tPos & 0xff);
                s.tPos >>= 8;
                if(s.rNToGo == 0){
                    s.rNToGo = BZ2_rNums[s.rTPos];
                    s.rTPos++;
                    if (s.rTPos == 512) s.rTPos = 0;
                }
                s.rNToGo--;
                //BZ_RAND_UPD_MASK;
                //#define BZ_RAND_UPD_MASK	       \
                //if (s->rNToGo == 0) {                       \
                //s->rNToGo = BZ2_rNums[s->rTPos];         \
                //s->rTPos++;                              \
                //if (s->rTPos == 512) s->rTPos = 0;       \
                //}                                           \
                //s->rNToGo--;
 
                k1 = (byte)((k1 & 0xff) ^ ((s.rNToGo == 1) ? 1 : 0));//BZ_RAND_MASK; 
                s.nblock_used++;
                if (s.nblock_used == s.save_nblock+1) continue;
                if ((k1 & 0xff) != s.k0) { s.k0 = (k1 & 0xff); continue; };
   
                s.state_out_len = 2;
                //BZ_GET_FAST(k1); 
                //BZ_RAND_UPD_MASK; 
                s.tPos = s.tt.getUInt32((int)s.tPos);//[(int)s.tPos];
                k1 = (byte)(s.tPos & 0xff);
                s.tPos >>= 8;
                if(s.rNToGo == 0){
                    s.rNToGo = BZ2_rNums[s.rTPos];
                    s.rTPos++;
                    if (s.rTPos == 512) s.rTPos = 0;
                }
                 s.rNToGo--;

		 k1 = (byte)((k1 & 0xff) ^ ((s.rNToGo == 1) ? 1 : 0));//BZ_RAND_MASK; 
                 s.nblock_used++;
                 if (s.nblock_used == s.save_nblock+1) continue;
                 if ((k1 & 0xff) != s.k0) { s.k0 = (k1 & 0xff); continue; };
   
                 s.state_out_len = 3;
                 //BZ_GET_FAST(k1); 
                 //BZ_RAND_UPD_MASK; 
                 s.tPos = s.tt.getUInt32((int)s.tPos);//[(int)s.tPos];
                 k1 = (byte)(s.tPos & 0xff);
                 s.tPos >>= 8;
                 if(s.rNToGo == 0){
                     s.rNToGo = BZ2_rNums[s.rTPos];
                     s.rTPos++;
                     if (s.rTPos == 512) s.rTPos = 0;
                 }
                 s.rNToGo--;

                 k1 = (byte)((k1 & 0xff) ^ ((s.rNToGo == 1) ? 1 : 0));//BZ_RAND_MASK; 
                 s.nblock_used++;
                 if (s.nblock_used == s.save_nblock+1) continue;
                 if ((k1 & 0xff) != s.k0) { s.k0 = (k1 & 0xff); continue; };
   
                 //BZ_GET_FAST(k1); 
                 //BZ_RAND_UPD_MASK; 
                 s.tPos = s.tt.getUInt32((int)s.tPos);//[(int)s.tPos];
                 k1 = (byte)(s.tPos & 0xff);
                 s.tPos >>= 8;
                 if(s.rNToGo == 0){
                     s.rNToGo = BZ2_rNums[s.rTPos];
                     s.rTPos++;
                     if (s.rTPos == 512) s.rTPos = 0;
                 }
                 s.rNToGo--;

                 k1 = (byte)((k1 & 0xff) ^ ((s.rNToGo == 1) ? 1 : 0));//BZ_RAND_MASK; 
                 s.nblock_used++;
                 s.state_out_len = (k1 & 0xff) + 4;

                 //BZ_GET_FAST(s.k0); 
                 //BZ_RAND_UPD_MASK; 
                 s.tPos = s.tt.getUInt32((int)s.tPos);//[(int)s.tPos];
                 k1 = (byte)(s.tPos & 0xff);
                 s.tPos >>= 8;
                 if(s.rNToGo == 0){
                     s.rNToGo = BZ2_rNums[s.rTPos];
                     s.rTPos++;
                     if (s.rTPos == 512) s.rTPos = 0;
                 }
                 s.rNToGo--;
                 s.k0 ^=  ((s.rNToGo == 1) ? 1 : 0);//BZ_RAND_MASK; 
                 s.nblock_used++;
            }
        } else {
            /* restore */
            long        c_calculatedBlockCRC = s.calculatedBlockCRC; //UInt32;
            byte         c_state_out_ch       = s.state_out_ch;
            int         c_state_out_len      = s.state_out_len;
            int         c_nblock_used        = s.nblock_used;
            int         c_k0                 = s.k0;
            IntView       c_tt                 = s.tt;
            long        c_tPos               = s.tPos; //UInt32
            ByteBuffer cs_next_out = s.strm.next_out;
            long  cs_avail_out         = s.strm.avail_out; //unsigned int?
            /* end restore */

            long       avail_out_INIT = cs_avail_out; //UInt32
            int        s_save_nblockPP = s.save_nblock+1;
            long total_out_lo32_old; //unsigned int?

            return_notr:
            while (true) {
                /* try to finish existing run */
                if (c_state_out_len > 0) {
                    while (true) {
	                if (cs_avail_out == 0) break return_notr;
                        if (c_state_out_len == 1) break;
                        //cs_next_out = c_state_out_ch;
                        cs_next_out.put(cs_next_out.position(), c_state_out_ch);
                        c_calculatedBlockCRC = BZ_UPDATE_CRC ( c_calculatedBlockCRC, c_state_out_ch );
                        c_state_out_len--;
                        //cs_next_out++;
                        cs_next_out.get();
                        cs_avail_out--;
                    }

                    //s_state_out_len_eq_one:            
                    {
                        if (cs_avail_out == 0) { 
                            c_state_out_len = 1; break return_notr;
                        }
                        //cs_next_out = c_state_out_ch;
                        cs_next_out.put(cs_next_out.position(), c_state_out_ch);
                        c_calculatedBlockCRC = BZ_UPDATE_CRC ( c_calculatedBlockCRC, c_state_out_ch );               
                        //cs_next_out++;
                        cs_next_out.get();
                        cs_avail_out--;
                    } 
	        }
                boolean s_state_out_len_eq_one_ctrl = false;
	        s_state_out_len_eq_one:
                do{
	            if(s_state_out_len_eq_one_ctrl){
                        s_state_out_len_eq_one_ctrl = false;
                        if (cs_avail_out == 0) { 
                            c_state_out_len = 1; break return_notr;
                        }
                        //cs_next_out = c_state_out_ch;
                        cs_next_out.put(cs_next_out.position(), c_state_out_ch);
                        c_calculatedBlockCRC = BZ_UPDATE_CRC ( c_calculatedBlockCRC, c_state_out_ch );
                        //cs_next_out++;
                        cs_next_out.get();
                        cs_avail_out--;
                    }

                    /* Only caused by corrupt data stream? */
                    if (c_nblock_used > s_save_nblockPP)
                       return true;

                    /* can a new run be started? */
                    if (c_nblock_used == s_save_nblockPP) {
                        c_state_out_len = 0;  
                        break return_notr;
                    }
                    c_state_out_ch = (byte)c_k0;
                    //BZ_GET_FAST_C(k1); 
                    //c_tPos = c_tt[c_tPos];	      \
                    //cccc = (UChar)(c_tPos & 0xff);            \
                    //c_tPos >>= 8;
                    c_tPos = c_tt.getUInt32((int)c_tPos);//[c_tPos];
                    k1 = (byte)(c_tPos & 0xff) ;
                    c_tPos >>= 8;
                    c_nblock_used++;
                    if ((k1 & 0xff) != c_k0) { 
                        c_k0 = k1 & 0xff; 
                        s_state_out_len_eq_one_ctrl = true;
                        continue s_state_out_len_eq_one;
                        //goto s_state_out_len_eq_one; 
                    }
                    if (c_nblock_used == s_save_nblockPP){
	                //goto s_state_out_len_eq_one;
                        s_state_out_len_eq_one_ctrl = true;
                        continue s_state_out_len_eq_one;
                    }
                }while(s_state_out_len_eq_one_ctrl);
   
                c_state_out_len = 2;
                //BZ_GET_FAST_C(k1); 
                c_tPos = c_tt.getUInt32((int)c_tPos);//[c_tPos];
                k1 = (byte)(c_tPos & 0xff);
                c_tPos >>= 8;

                c_nblock_used++;
                if (c_nblock_used == s_save_nblockPP) continue;
                if ((k1 & 0xff) != c_k0) { 
                    c_k0 = (k1 & 0xff); 
                    continue; 
                };
   
                c_state_out_len = 3;
                //BZ_GET_FAST_C(k1); 
                c_tPos = c_tt.getUInt32((int)c_tPos);//[c_tPos];
                k1 = (byte)(c_tPos & 0xff);
                c_tPos >>= 8;

                c_nblock_used++;
                if (c_nblock_used == s_save_nblockPP) continue;
                if ((k1 & 0xff) != c_k0) { 
                    c_k0 = (k1 & 0xff); 
                    continue; 
                };
   
                //BZ_GET_FAST_C(k1); 
                c_tPos = c_tt.getUInt32((int)c_tPos);//[c_tPos];
                k1 = (byte)(c_tPos & 0xff);
                c_tPos >>= 8;

                c_nblock_used++;
                c_state_out_len = (k1 & 0xff) + 4;        
                //BZ_GET_FAST_C(c_k0); 
                c_tPos = c_tt.getUInt32((int)c_tPos);//[c_tPos];
                c_k0 = ((byte)(c_tPos & 0xff)) & 0xff;
                c_tPos >>= 8;
                c_nblock_used++;
            }

            //return_notr:
            total_out_lo32_old = s.strm.total_out_lo32;
            s.strm.total_out_lo32 += (avail_out_INIT - cs_avail_out);      
            if (s.strm.total_out_lo32 < total_out_lo32_old)
                s.strm.total_out_hi32++;

            /* save */
            s.calculatedBlockCRC = c_calculatedBlockCRC;
            s.state_out_ch       = c_state_out_ch;
            s.state_out_len      = c_state_out_len;
            s.nblock_used        = c_nblock_used;
            s.k0                 = c_k0;
            s.tt                 = c_tt;
            s.tPos               = c_tPos;
            //s.strm.next_out     = cs_next_out;
            //s.strm.put_next_out(cs_next_out);
            s.strm.avail_out    = cs_avail_out;
            /* end save */
        }
        return false;
    }


    public static boolean unRLE_obuf_to_output_SMALL(DState s){
        //UChar k1; 
        byte k1;
        if (s.blockRandomised) {
            while (true) {
                /* try to finish existing run */
                while (true) {
                    if (s.strm.avail_out == 0) return false;
                    if (s.state_out_len == 0) break;
                    //*( (UChar*)(s->strm->next_out) ) = s->state_out_ch;
                    s.strm.put_next_out(s.state_out_ch);
                    s.calculatedBlockCRC = BZ_UPDATE_CRC ( s.calculatedBlockCRC, s.state_out_ch );
                    s.state_out_len--;
                    //s.strm.next_out++;
                    s.strm.inc_next_out();
                    s.strm.avail_out--;
                    s.strm.total_out_lo32++;
                    if (s.strm.total_out_lo32 == 0) s.strm.total_out_hi32++;
                }
   
                /* can a new run be started? */
                if (s.nblock_used == s.save_nblock+1) return false;

                /* Only caused by corrupt data stream? */
                if (s.nblock_used > s.save_nblock+1)
	            return true;;
   
                s.state_out_len = 1;
                s.state_out_ch = (byte)s.k0;

                //BZ_GET_SMALL(k1); 
                //#define BZ_GET_SMALL(cccc)		      \
                //cccc = BZ2_indexIntoF ( s->tPos, s->cftab );    \
                //s->tPos = GET_LL(s->tPos);
                //#define GET_LL(i)			\
                //(((UInt32)s->ll16[i]) | (GET_LL4(i) << 16))
                //#define GET_LL4(i)					\
                //((((UInt32)(s->ll4[(i) >> 1])) >> (((i) << 2) & 0x4)) & 0xF)

                k1 = (byte)BZ2_indexIntoF((int)s.tPos, s.cftab);
                s.tPos = (s.ll16.getUInt16((int)s.tPos)) | 
                         ((((((s.ll4[(int)(s.tPos) >> 1])) >> ((((s.tPos) << 2) & UINT32) & 0x4)) & 0xF) << 16) & UINT32);
                if(s.rNToGo == 0) {
                    s.rNToGo = BZ2_rNums[s.rTPos];
                    s.rTPos++;
	            if (s.rTPos == 512) s.rTPos = 0;
                }
                s.rNToGo--;
                //BZ_RAND_UPD_MASK;
                //#define BZ_RAND_UPD_MASK		       \
                //if (s->rNToGo == 0) {                       \
                //s->rNToGo = BZ2_rNums[s->rTPos];         \
                //s->rTPos++;                              \
                //if (s->rTPos == 512) s->rTPos = 0;       \
                //}                                           \
                //s->rNToGo--;

                k1 = (byte)((k1 & 0xff) ^ ((s.rNToGo == 1) ? 1 : 0));//BZ_RAND_MASK; 
                s.nblock_used++;
                if (s.nblock_used == s.save_nblock+1) continue;
                if ((k1 & 0xff) != s.k0) { s.k0 = (k1 & 0xff); continue; };
   
                s.state_out_len = 2;

                //BZ_GET_SMALL(k1); 
                //BZ_RAND_UPD_MASK; 
                k1 = (byte)BZ2_indexIntoF((int)s.tPos, s.cftab);
                s.tPos = (s.ll16.getUInt16((int)s.tPos)) | 
                         ((((((s.ll4[(int)(s.tPos) >> 1])) >> ((((s.tPos) << 2) & UINT32) & 0x4)) & 0xF) << 16) & UINT32);
                if(s.rNToGo == 0) {
                    s.rNToGo = BZ2_rNums[s.rTPos];
                    s.rTPos++;
	            if (s.rTPos == 512) s.rTPos = 0;
                }
                s.rNToGo--;

                k1 = (byte)((k1 & 0xff) ^ ((s.rNToGo == 1) ? 1 : 0));//BZ_RAND_MASK; 
                s.nblock_used++;
                if (s.nblock_used == s.save_nblock+1) continue;
                if ((k1 & 0xff) != s.k0) { s.k0 = (k1 & 0xff); continue; };
   
                s.state_out_len = 3;
                //BZ_GET_SMALL(k1); 
                //BZ_RAND_UPD_MASK; 
                k1 = (byte)BZ2_indexIntoF((int)s.tPos, s.cftab);
                s.tPos = (s.ll16.getUInt16((int)s.tPos)) | 
                         ((((((s.ll4[(int)(s.tPos) >> 1])) >> ((((s.tPos) << 2) & UINT32) & 0x4)) & 0xF) << 16) & UINT32);
                if(s.rNToGo == 0) {
                    s.rNToGo = BZ2_rNums[s.rTPos];
                    s.rTPos++;
	            if (s.rTPos == 512) s.rTPos = 0;
                }
                s.rNToGo--;

                k1 = (byte)((k1 & 0xff) ^ ((s.rNToGo == 1) ? 1 : 0));//BZ_RAND_MASK; 
                s.nblock_used++;
                if (s.nblock_used == s.save_nblock+1) continue;
                if ((k1 & 0xff) != s.k0) { s.k0 = (k1 & 0xff); continue; };
   
                //BZ_GET_SMALL(k1); 
                //BZ_RAND_UPD_MASK;
                k1 = (byte)BZ2_indexIntoF((int)s.tPos, s.cftab);
                s.tPos = (s.ll16.getUInt16((int)s.tPos)) | 
                         ((((((s.ll4[(int)(s.tPos) >> 1])) >> ((((s.tPos) << 2) & UINT32) & 0x4)) & 0xF) << 16) & UINT32);
                if(s.rNToGo == 0) {
                    s.rNToGo = BZ2_rNums[s.rTPos];
                    s.rTPos++;
	            if (s.rTPos == 512) s.rTPos = 0;
                }
                s.rNToGo--;

                k1 = (byte)((k1 & 0xff) ^ ((s.rNToGo == 1) ? 1 : 0));//BZ_RAND_MASK; 
                s.nblock_used++;
                s.state_out_len = (k1 & 0xff) + 4;

                //BZ_GET_SMALL(s.k0); 
                //BZ_RAND_UPD_MASK;
                s.k0 = BZ2_indexIntoF((int)s.tPos, s.cftab);
                s.tPos = (s.ll16.getUInt16((int)s.tPos)) | 
                         ((((((s.ll4[(int)(s.tPos) >> 1])) >> ((((s.tPos) << 2) & UINT32) & 0x4)) & 0xF) << 16) & UINT32);
                if(s.rNToGo == 0) {
                    s.rNToGo = BZ2_rNums[s.rTPos];
                    s.rTPos++;
	            if (s.rTPos == 512) s.rTPos = 0;
                }
                s.rNToGo--;

                s.k0 ^= ((s.rNToGo == 1) ? 1 : 0);//BZ_RAND_MASK; 
                s.nblock_used++;
            }
        } else {
            while (true) {
                /* try to finish existing run */
                while (true) {
                    if (s.strm.avail_out == 0) return false;
                    if (s.state_out_len == 0) break;
                    //*( (UChar*)(s->strm->next_out) ) = s->state_out_ch;
                    s.strm.put_next_out(s.state_out_ch);
                    s.calculatedBlockCRC = BZ_UPDATE_CRC (s.calculatedBlockCRC, s.state_out_ch );
                    s.state_out_len--;
                    //s.strm.next_out++;
	            s.strm.inc_next_out();
                    s.strm.avail_out--;
                    s.strm.total_out_lo32++;
                    if (s.strm.total_out_lo32 == 0) s.strm.total_out_hi32++;
                }
   
                /* can a new run be started? */
                if (s.nblock_used == s.save_nblock+1) return false;

                /* Only caused by corrupt data stream? */
                if (s.nblock_used > s.save_nblock+1)
                    return true;
   
                s.state_out_len = 1;
                s.state_out_ch = (byte)s.k0;
                //BZ_GET_SMALL(k1); 
                k1 = (byte)BZ2_indexIntoF((int)s.tPos, s.cftab);
                s.tPos = (s.ll16.getUInt16((int)s.tPos)) | 
	                 ((((((s.ll4[(int)(s.tPos) >> 1])) >> ((((s.tPos) << 2) & UINT32) & 0x4)) & 0xF) << 16) & UINT32);
                s.nblock_used++;
                if (s.nblock_used == s.save_nblock+1) continue;
                if ((k1 & 0xff) != s.k0) { s.k0 = (k1 & 0xff); continue; };
   
                s.state_out_len = 2;
                //BZ_GET_SMALL(k1); 
                k1 = (byte)BZ2_indexIntoF((int)s.tPos, s.cftab);
                s.tPos = (s.ll16.getUInt16((int)s.tPos)) | 
                         ((((((s.ll4[(int)(s.tPos) >> 1])) >> ((((s.tPos) << 2) & UINT32) & 0x4)) & 0xF) << 16) & UINT32);
                s.nblock_used++;
                if (s.nblock_used == s.save_nblock+1) continue;
                if ((k1 & 0xff) != s.k0) { s.k0 = (k1 & 0xff); continue; };
   
                s.state_out_len = 3;
                //BZ_GET_SMALL(k1); 
                k1 = (byte)BZ2_indexIntoF((int)s.tPos, s.cftab);
                s.tPos = (s.ll16.getUInt16((int)s.tPos)) | 
                         ((((((s.ll4[(int)(s.tPos) >> 1])) >> ((((s.tPos) << 2) & UINT32) & 0x4)) & 0xF) << 16) & UINT32);
                s.nblock_used++;
                if (s.nblock_used == s.save_nblock+1) continue;
                if ((k1 & 0xff) != s.k0) { s.k0 = (k1 & 0xff); continue; };
   
                //BZ_GET_SMALL(k1); 
                k1 = (byte)BZ2_indexIntoF((int)s.tPos, s.cftab);
                s.tPos = (s.ll16.getUInt16((int)s.tPos)) | 
                         ((((((s.ll4[(int)(s.tPos) >> 1])) >> ((((s.tPos) << 2) & UINT32) & 0x4)) & 0xF) << 16) & UINT32);
                s.nblock_used++;
                s.state_out_len = (k1 & 0xff) + 4;
                //BZ_GET_SMALL(s.k0); 
                s.k0 = BZ2_indexIntoF((int)s.tPos, s.cftab);
                s.tPos = (s.ll16.getUInt16((int)s.tPos)) | 
	              ((((((s.ll4[(int)(s.tPos) >> 1])) >> ((((s.tPos) << 2) & UINT32) & 0x4)) & 0xF) << 16) & UINT32);
                s.nblock_used++;
            }
        }
    }

    public static int BZ2_bzDecompress(bz_stream strm){
        boolean    corrupt;
        DState s;
        if (strm == null) return BZ_PARAM_ERROR;
        s = (DState)strm.state;
        if (s == null) return BZ_PARAM_ERROR;
        if (s.strm != strm) return BZ_PARAM_ERROR;

        while (true) {
            if (s.state == BZ_X_IDLE) return BZ_SEQUENCE_ERROR;
            if (s.state == BZ_X_OUTPUT) {
                if (s.smallDecompress)
                    corrupt = unRLE_obuf_to_output_SMALL ( s ); 
                else
                    corrupt = unRLE_obuf_to_output_FAST  ( s );
                if (corrupt) return BZ_DATA_ERROR;
                if (s.nblock_used == s.save_nblock+1 && s.state_out_len == 0) {
                    s.calculatedBlockCRC = (~s.calculatedBlockCRC) & UINT32;
                    //if (s->verbosity >= 3) 
	            //  VPrintf2 ( " {0x%08x, 0x%08x}", s->storedBlockCRC, 
	            //            s->calculatedBlockCRC );
                    //if (s->verbosity >= 2) VPrintf0 ( "]" );
	            if (s.calculatedBlockCRC != s.storedBlockCRC){
                        return BZ_DATA_ERROR;
                    }
                    s.calculatedCombinedCRC 
		        = ((s.calculatedCombinedCRC << 1) & UINT32) | 
                            (s.calculatedCombinedCRC >> 31);
                    s.calculatedCombinedCRC =  (s.calculatedCombinedCRC ^ s.calculatedBlockCRC) & UINT32;
                    s.state = BZ_X_BLKHDR_1;
                 } else {
                     return BZ_OK;
                 }
            }
            if (s.state >= BZ_X_MAGIC_1) {
                int r = decompress.BZ2_decompress ( s );
                if (r == BZ_STREAM_END) {
	            //if (s->verbosity >= 3)
	            //VPrintf2 ( "\n    combined CRCs: stored = 0x%08x, computed = 0x%08x", 
	            //           s->storedCombinedCRC, s->calculatedCombinedCRC );
                   if (s.calculatedCombinedCRC != s.storedCombinedCRC)
                       return BZ_DATA_ERROR;
                   return r;
                }
                if (s.state != BZ_X_OUTPUT) return r;
            }
        }
        //AssertH ( 0, 6001 );
    }


    public static int BZ2_bzDecompressEnd(bz_stream strm){
        DState s;
        if (strm == null) return BZ_PARAM_ERROR;
        s = (DState)strm.state;
        if (s == null) return BZ_PARAM_ERROR;
        if (s.strm != strm) return BZ_PARAM_ERROR;

        s.tt = null; s.ll16 = null; s.ll4 = null;
        strm.state = null;
        return BZ_OK;
    }


    public static bzFile BZ2_bzReadOpen
                       ( int[]  bzerror, 
                         ReadableByteChannel f, 
                         int   verbosity,
                         boolean   small,
                         byte[] unused,
                         long[]   nUnused ){
        bzFile bzf = null;
        int     ret;

        BZ_SETERR(bzf, BZ_OK, bzerror);

        if (f == null || 
         (verbosity < 0 || verbosity > 4) ||
         (unused == null && nUnused[0] != 0) ||
         (unused != null && (nUnused[0] < 0 || nUnused[0] > BZ_MAX_UNUSED)))
	    { BZ_SETERR(bzf, BZ_PARAM_ERROR, bzerror); return null; };

        //if (ferror(f))
        // { BZ_SETERR(BZ_IO_ERROR); return NULL; };

        bzf = new bzFile();
        BZ_SETERR(bzf, BZ_OK, bzerror);

        bzf.initialisedOk = false;
        bzf.handle        = f;
        bzf.bufN          = 0;
        bzf.writing       = false;
        bzf.strm.bzalloc  = null;
        bzf.strm.bzfree   = null;
        bzf.strm.opaque   = null;
   
        //while (nUnused > 0) {
        // bzf.buf[bzf.bufN] = *((UChar*)(unused)); bzf->bufN++;
        // unused = ((void*)( 1 + ((UChar*)(unused))  ));
        // nUnused--;
        //}

        ret = BZ2_bzDecompressInit(bzf.strm, verbosity, small );
        if (ret != BZ_OK)
           { BZ_SETERR(bzf, ret, bzerror); //free(bzf); 
             return null; };

        bzf.strm.avail_in = bzf.bufN;
        bzf.strm.next_in  = bzf.buf;

        bzf.initialisedOk = true;
        return bzf;   
    }

    /*---------------------------------------------------*/
    public static void BZ2_bzReadClose (int[] bzerror, bzFile b ){
        bzFile bzf = b;

        BZ_SETERR(bzf, BZ_OK, bzerror);
        if (bzf == null)
           { BZ_SETERR(bzf, BZ_OK, bzerror); return; };

        if (bzf.writing)
           { BZ_SETERR(bzf, BZ_SEQUENCE_ERROR, bzerror); return; };

        if (bzf.initialisedOk)
            BZ2_bzDecompressEnd (bzf.strm);
        //free ( bzf );
    }


    /*---------------------------------------------------*/
    public static int BZ2_bzRead 
               ( int[]    bzerror, 
                 bzFile b, 
                 ByteBuffer   buf, 
                 int     len )
    {
        int   n, ret;
        bzFile bzf = b;

        BZ_SETERR(bzf, BZ_OK, bzerror);

        if (bzf == null || buf == null || len < 0)
            { BZ_SETERR(bzf, BZ_PARAM_ERROR, bzerror); return 0; };

        if (bzf.writing)
            { BZ_SETERR(bzf, BZ_SEQUENCE_ERROR, bzerror); return 0; };

        if (len == 0)
            { BZ_SETERR(bzf, BZ_OK, bzerror); return 0; };

        bzf.strm.avail_out = len;
        bzf.strm.next_out = buf;
        ReadableByteChannel handle = (ReadableByteChannel)bzf.handle;
        while (true) {

           //if (ferror(bzf->handle)) 
           //{ BZ_SETERR(BZ_IO_ERROR); return 0; };

           if (bzf.strm.avail_in == 0){// && !myfeof(bzf->handle)) {
	       //n = fread ( bzf->buf, sizeof(UChar), 
	       //        BZ_MAX_UNUSED, bzf->handle );
	       try{
                   bzf.buf.clear();
	           n = handle.read(bzf.buf);
                   bzf.buf.flip();
	       } catch(Exception x){ 
                   x.printStackTrace();
	           BZ_SETERR(bzf, BZ_IO_ERROR, bzerror);
                   return 0; 
               }
               //if (ferror(bzf->handle))
	       // { BZ_SETERR(BZ_IO_ERROR); return 0; };
               bzf.bufN = n;
               bzf.strm.avail_in = bzf.bufN;
               bzf.strm.next_in = bzf.buf;
           }

           ret = BZ2_bzDecompress(bzf.strm);
      
           if (ret != BZ_OK && ret != BZ_STREAM_END){ 
               BZ_SETERR(bzf, ret, bzerror); 
               return 0; 
           }

           //if (ret == BZ_OK && myfeof(bzf->handle) && 
           //  bzf.strm.avail_in == 0 && bzf.strm.avail_out > 0)
           // { BZ_SETERR(BZ_UNEXPECTED_EOF); return 0; };

           if (ret == BZ_STREAM_END){ 
               BZ_SETERR(bzf, BZ_STREAM_END, bzerror);
	       return (int)(len - bzf.strm.avail_out); 
           }
           if (bzf.strm.avail_out == 0)
	   { BZ_SETERR(bzf, BZ_OK, bzerror); return len; };
              
        }
    }

    /*---------------------------------------------------*/
    public static void BZ2_bzReadGetUnused 
                         ( int[]    bzerror, 
                           bzFile b, 
                           ByteBuffer[]  unused, 
                           long[]    nUnused )
    {
       bzFile bzf = b;
       if (bzf == null)
           { BZ_SETERR(null, BZ_PARAM_ERROR, bzerror); return; };
       if (bzf.lastErr != BZ_STREAM_END)
           { BZ_SETERR(b, BZ_SEQUENCE_ERROR, bzerror); return; };
       if (unused == null || nUnused == null)
           { BZ_SETERR(b, BZ_PARAM_ERROR, bzerror); return; };

       BZ_SETERR(b, BZ_OK, bzerror);
       nUnused[0] = bzf.strm.avail_in;
       unused[0] = bzf.strm.next_in.slice();//bzf.strm.get_next_in();//bzf.strm.next_in;
    }


    /*---------------------------------------------------*/
    public static int BZ2_bzread (bzFile b, ByteBuffer buf, int len ){
        int[] bzerr = new int[1]; 
        int nread;
        if (b.lastErr == BZ_STREAM_END) return 0;
        nread = BZ2_bzRead(bzerr,b,buf,len);
        if (bzerr[0] == BZ_OK || bzerr[0] == BZ_STREAM_END) {
            return nread;
        } else {
            return -1;
        }
    }

    /*---------------------------------------------------*/
    public static int BZ2_bzwrite(bzFile b, ByteBuffer buf, int len ) throws IOException{
        int[] bzerr = new int[1];
        BZ2_bzWrite(bzerr,b,buf,len);
        if(bzerr[0] == BZ_OK){
            return len;
        }else{
            return -1;
        }
    }

    public static void BZ_SETERR(bzFile bzf, int eee){
        BZ_SETERR(bzf, eee, null);
    }

    public static void BZ_SETERR(bzFile bzf, int eee, int[] bzerror){
        if(bzerror != null) bzerror[0] = eee;
        if(bzf != null) bzf.lastErr = eee;    
    }

    public static String BZ2_bzlibVersion(){
        return BZ_VERSION;
    }

    /*---------------------------------------------------*/
    public static int BZ2_bzBuffToBuffCompress
                             ( ByteBuffer         dest, 
                               int[] destLen,
                               ByteBuffer         source, 
                               long  sourceLen,
                               int           blockSize100k, 
                               int           verbosity, 
                               int           workFactor ){
        bz_stream strm = new bz_stream();
        int ret;

        if (dest == null || destLen == null || 
           source == null ||
           blockSize100k < 1 || blockSize100k > 9 ||
           verbosity < 0 || verbosity > 4 ||
           workFactor < 0 || workFactor > 250) 
            return BZ_PARAM_ERROR;

        if (workFactor == 0) workFactor = 30;
            strm.bzalloc = null;
            strm.bzfree = null;
            strm.opaque = null;
            ret = BZ2_bzCompressInit ( strm, blockSize100k, 
                                  verbosity, workFactor );
        if (ret != BZ_OK) return ret;

        strm.next_in = source;
        strm._next_in = source.array(); 
        strm._ni = source.position();  
        strm.next_out = dest;
        strm.avail_in = sourceLen;
        strm.avail_out = destLen[0];
   
        errhandler: {
        output_overflow:{
        ret = BZ2_bzCompress (strm, BZ_FINISH );
        if (ret == BZ_FINISH_OK) break output_overflow;
        if (ret != BZ_STREAM_END) break errhandler;

        /* normal termination */
        destLen[0] -= strm.avail_out;   
        BZ2_bzCompressEnd ( strm );
        return BZ_OK;

        //output_overflow:
        }
        BZ2_bzCompressEnd (strm );
        return BZ_OUTBUFF_FULL;

        //errhandler:
        }
        BZ2_bzCompressEnd (strm );
        return ret;
    }


    /*---------------------------------------------------*/
    public static int BZ2_bzBuffToBuffDecompress
                               ( ByteBuffer         dest, 
                                 int[] destLen,
                                 ByteBuffer         source, 
                                 long  sourceLen,
                                 int           small,
                                 int           verbosity ){
        bz_stream strm = new bz_stream();
        int ret;

        if (dest == null || destLen == null || 
            source == null ||
            (small != 0 && small != 1) ||
             verbosity < 0 || verbosity > 4) 
             return BZ_PARAM_ERROR;

        strm.bzalloc = null;
        strm.bzfree = null;
        strm.opaque = null;
        ret = BZ2_bzDecompressInit (strm, verbosity, small != 0 );
        if (ret != BZ_OK) return ret;

        strm.next_in = source;
        strm._next_in = source.array();
        strm._ni = source.position();
        strm.next_out = dest;
        strm.avail_in = sourceLen;
        strm.avail_out = destLen[0];

        ret = BZ2_bzDecompress (strm );
        errhandler:{
        output_overflow_or_eof:{

        if (ret == BZ_OK) break output_overflow_or_eof;
        if (ret != BZ_STREAM_END) break errhandler;

        /* normal termination */
        destLen[0] -= strm.avail_out;
        BZ2_bzDecompressEnd (strm );
        return BZ_OK;

       //output_overflow_or_eof:
       }
       if (strm.avail_out > 0) {
          BZ2_bzDecompressEnd ( strm );
          return BZ_UNEXPECTED_EOF;
       } else {
          BZ2_bzDecompressEnd ( strm );
          return BZ_OUTBUFF_FULL;
       }

       }
       //errhandler:
       BZ2_bzDecompressEnd ( strm );
       return ret; 
    }


} 