package org.jythonx.bzip2;

import java.io.*;
import java.nio.*;
import java.nio.channels.*;

public class unzcrash implements BZConstants{

    static final int M_BLOCK = 1000000;
    static final int M_BLOCK_OUT = (M_BLOCK + 1000000);

    ByteBuffer inbuf = ByteBuffer.allocate(M_BLOCK);
    ByteBuffer outbuf = ByteBuffer.allocate(M_BLOCK_OUT);
    ByteBuffer zbuf = ByteBuffer.allocate(M_BLOCK + 600 + (M_BLOCK / 100));
    byte[] zbuf_i = zbuf.array();

    int nIn;
    int[] nOut = new int[1];
    int[] nZ = new int[1];

    static final String[] bzerrorstrings = {
       "OK"
      ,"SEQUENCE_ERROR"
      ,"PARAM_ERROR"
      ,"MEM_ERROR"
      ,"DATA_ERROR"
      ,"DATA_ERROR_MAGIC"
      ,"IO_ERROR"
      ,"UNEXPECTED_EOF"
      ,"OUTBUFF_FULL"
      ,"???"   /* for future */
      ,"???"   /* for future */
      ,"???"   /* for future */
      ,"???"   /* for future */
      ,"???"   /* for future */
      ,"???"   /* for future */
    };

    public unzcrash(String[] argv) throws IOException{
        ReadableByteChannel f = null;
        int   r;
        int   bit;
        int   i;

        if (argv.length != 1) {
            System.err.println("usage: unzcrash filename");
            return;
        }

        //f = fopen ( argv[1], "r" );
        try{
            f = new RandomAccessFile(argv[0], "r").getChannel();
        } catch(IOException x){}
        if (f == null) {
           //fprintf ( stderr, "unzcrash: can't open %s\n", argv[1] );
           System.err.println("unzcrash: can't open " + argv[0]);
          return;
        }

        //nIn = fread ( inbuf, 1, M_BLOCK, f );
        inbuf.position(0);
        inbuf.limit(M_BLOCK);
        nIn = f.read(inbuf);
        inbuf.flip();
        //fprintf ( stderr, "%d bytes read\n", nIn );
        System.err.println(nIn + " bytes read");

        nZ[0] = M_BLOCK;
        r = BZLib.BZ2_bzBuffToBuffCompress (zbuf, nZ, inbuf, (long)nIn, 9, 0, 30 );
        assert (r == BZ_OK);
        //fprintf ( stderr, "%d after compression\n", nZ );
        System.err.println(nZ[0] + " after compression");
        for (bit = 0; bit < nZ[0]*8; bit++) {
            //fprintf ( stderr, "bit %d  ", bit );
            System.err.print("bit " + bit + "  ");
            flip_bit ( bit );
            nOut[0] = M_BLOCK_OUT;
            zbuf.position(0);
            outbuf.clear();
            r = BZLib.BZ2_bzBuffToBuffDecompress (outbuf, nOut, zbuf, nZ[0], 0, 0 );
            //fprintf ( stderr, " %d  %s ", r, bzerrorstrings[-r] );
            System.err.print(" " + r + "  " + bzerrorstrings[-r] + " ");

            if (r != BZ_OK) {
	        //fprintf ( stderr, "\n" );
                System.err.println("");
            } else {
                if (nOut[0] != nIn) {
	            //fprintf(stderr, "nIn/nOut mismatch %d %d\n", nIn, nOut );
	            System.err.println("nIn/nOut mismatch " + nIn + " " + nOut);
                    return;
                } else {
                    for (i = 0; i < nOut[0]; i++)
	                if (inbuf.get(i) != outbuf.get(i)) { 
		            //fprintf(stderr, "mismatch at %d\n", i ); 
		            System.err.println("mismatch at " + i);
                            return; 
                         }
                         if (i == nOut[0])
	                     // fprintf(stderr, "really ok!\n" );
                             System.err.println("really ok!");
                }
            }
            flip_bit ( bit );
       }

       //#if 0
       assert (nOut[0] == nIn);
       byte[] in_arr = inbuf.array();
       byte[] out_arr = outbuf.array();
       for (i = 0; i < nOut[0]; i++) {
           //if (inbuf.get(i) != outbuf.get(i)){
           if(in_arr[i] != out_arr[i]){
	       //fprintf ( stderr, "difference at %d !\n", i );
               System.err.println("difference at " + i + " !");
               return;
               //return 1;
           }
        }
       //#endif
       //fprintf ( stderr, "all ok\n" );
       System.err.println("all ok");
       return;
    }

    void flip_bit (int bit){
       int byteno = bit / 8;
       int bitno  = bit % 8;
       int mask = (1 << bitno);
       //fprintf ( stderr, "(byte %d  bit %d  mask %d)",
       //          byteno, bitno, (int)mask );
       zbuf_i[byteno] = (byte)((zbuf_i[byteno] & 0xff) ^ mask);
    }

    public static void main (String[] argv) throws IOException{
	new unzcrash(argv);
    }


}