package org.jythonx.modules;

import org.python.core.*;
import org.jythonx.utils.*;

public class itertools{

    public static PyType chain = ChainIterator.type;
    public static PyType imap = ImapIterator.type;
    public static PyType count = CountIterator.type;
    public static PyType islice = IsliceIterator.type;
    public static PyType izip = IzipIterator.type;
    public static PyType repeat = RepeatIterator.type;
    public static PyType cycle = CycleIterator.type;
    public static PyType groupby = GroupbyIterator.type;
    public static PyType dropwhile = DropwhileIterator.type;
    public static PyType starmap = StarmapIterator.type;
    public static PyType takewhile = TakewhileIterator.type;
    public static PyType ifilter = IfilterIterator.type;
    public static PyType ifilterfalse = IfilterfalseIterator.type;
    public static PyBuiltinFunction tee;
    static{ 
	PyBuiltinFunction.DefaultInfo info = new PyBuiltinFunction.DefaultInfo("tee", 1,2);
        tee = new Tee(null, info);
    }

    public static class ChainIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(ChainIterator.class);
        public static final String exposed_name = "itertools.chain";

        public static void typeSetup(PyObject dict,PyType.Newstyle marker) {
             dict.__setitem__("__new__",new PyNewWrapper(ChainIterator.class,"__new__",-1,-1) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     return chain_new(subtype,args,keywords);
                 }
            });
        }

        public static PyObject chain_new(PyType type, PyObject[] iterables, String[] kwords){
            if(type == ChainIterator.type){
                MacroEmulation.PyArg_NoKeywords("chain()", kwords);
             }

	    PyObject[] iterators = new PyObject[iterables.length];
            for(int i = 0; i < iterables.length; i ++){
                try{
                    iterators[i] = iterables[i].__iter__();
                } catch(PyException x){
		    if(Py.matchException(x, Py.TypeError)){
                        throw Py.TypeError("chain argument #" 
                                           +(i+1)+ 
                                           " must support iteration");
                    }
                    throw x;
		}
            }
            return new ChainIterator(type, iterators);
        }

        PyObject[] iterators;
        int iternum;
        ChainIterator(PyType type, PyObject[] iterators){
            super(type);
            this.iterators = iterators;
        }

	public PyObject __iternext__(){
	    PyObject it;
	    PyObject item;

	    while (iternum < iterators.length) {
	        it = iterators[iternum];
		item = PyFunctions.PyIter_Next(it);
		if (item != null)
		    return item;
	        iternum++;
	    }
	    return null;
        }
    }

    public static class ImapIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(ImapIterator.class);
        public static final String exposed_name = "itertools.imap";
        public static void typeSetup(PyObject dict,PyType.Newstyle marker) {
             dict.__setitem__("__new__",new PyNewWrapper(ImapIterator.class,"__new__",-1,-1) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     return imap_new(subtype,args,keywords);
                 }
            });
        }


        static PyObject imap_new(PyType type, PyObject[] args, String[] kwds){
	    PyObject it, func;

	    if (type == ImapIterator.type){
	        MacroEmulation.PyArg_NoKeywords("imap()", kwds);
            }

	    if(args.length < 2) {
                throw Py.TypeError("imap() must have at least two arguments.");
	    }

            PyObject[] iters = new PyObject[args.length -1];
	    for (int i=1 ; i< args.length ; i++) {
                it = args[i].__iter__();
                iters[i - 1] = it;
	    }

            func = args[0];
	    return new ImapIterator(type, func, iters);
         }


	 PyObject func;
	 PyObject[] iters;
	 ImapIterator(PyType type, PyObject func, PyObject[] iters){
	     super(type);
             this.func = func;
             this.iters = iters;
         }

        public PyObject __iternext__(){
	    PyObject val;
            int numargs = iters.length;
	    PyObject[] argtuple = new PyObject[numargs];

	    for (int i=0 ; i<numargs ; i++) {
		val = PyFunctions.PyIter_Next(iters[i]);
		if (val == null) {
		    return null;
		}
                argtuple[i] = val;
	    }
	    if(func == Py.None || func == null)
                return new PyTuple(argtuple);
	    return func.__call__(argtuple, Py.NoKeywords);
        }
    }


    public static class CountIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(CountIterator.class);
        public static final String exposed_name = "itertools.count";
        public static void typeSetup(PyObject dict,PyType.Newstyle marker) {
             dict.__setitem__("__new__",new PyNewWrapper(CountIterator.class,"__new__",0,1) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     if(!DefaultInfo.check(args.length, info.getMinargs(), info.getMaxargs())){
                         throw info.unexpectedCall(args.length, false);
		     }
                     return count_new(subtype,args,keywords);
                 }
            });
        }


        static PyObject count_new(PyType type, PyObject[] args, String[] kwds){
	    if (type == CountIterator.type){
	        MacroEmulation.PyArg_NoKeywords("count()", kwds);
            }

            PyObject cnt = Py.Zero;
            if(args.length > 0){
                boolean throwTypeError = false;
                try{
                    cnt = args[0].__int__();
                    if(!(cnt instanceof PyInteger) && !(cnt instanceof PyLong))
                        throwTypeError = true;
		} catch(PyException x){ throwTypeError = true; }
                if(throwTypeError)
                    throw Py.TypeError("an integer is required");
            }

	    return new CountIterator(type, cnt);
         }

	 PyObject cnt;
	 CountIterator(PyType type, PyObject cnt){
	     super(type);
             this.cnt = cnt;
         }

        public PyObject __iternext__(){
            if(cnt instanceof PyLong && ((PyLong)cnt).getValue().longValue() == Long.MAX_VALUE){
                throw Py.OverflowError("cannot count beyond LONG_MAX");
            }
	    PyObject rv = cnt;
            cnt = cnt.__add__(Py.One);
            return rv;
        }

        public PyString __repr__(){
            return count_repr();
        }
 
        public PyString count_repr(){
	    PyString fmt = new PyString("count(%d)");
	    return (PyString)fmt.__mod__(cnt);
        }

    }

    public static class IsliceIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(IsliceIterator.class);
        public static final String exposed_name = "itertools.islice";

        public static void typeSetup(PyObject dict,PyType.Newstyle marker) {
             dict.__setitem__("__new__",new PyNewWrapper(IsliceIterator.class,"__new__",2,4) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     if(!DefaultInfo.check(args.length, info.getMinargs(), info.getMaxargs())){
                         throw info.unexpectedCall(args.length, false);
		     }
                     return islice_new(subtype,args,keywords);
                 }
            });
        }


        static PyObject islice_new(PyType type, PyObject[] args, String[] kwds){
	    if (type == IsliceIterator.type){
	        MacroEmulation.PyArg_NoKeywords("islice()", kwds);
            }
            PyObject seq = args[0];
	    int start= 0, stop= -1, step = 1;

            int numargs = args.length;
	    PyObject a1 = args[1];
            PyObject a2 = numargs > 2? args[2]: null;
            PyObject a3 = numargs > 3? args[3]: null;

	    if (numargs == 2) {
		if (a1 != Py.None) {
                    boolean throwValueError  =false;
                    try{
		        PyObject inter = a1.__int__();
                        if(inter instanceof PyInteger)
                            stop = ((PyInteger)inter).getValue();
                        else if(inter instanceof PyLong)
                            stop = ((PyLong)inter).getValue().intValue();
                        else
                            throwValueError = true;
		    } catch(PyException x){
                        throwValueError = true;
                    }
                    if(throwValueError)
                        throw Py.ValueError("Stop argument for islice() must be a non-negative integer or None.");
		}
	    } else {
		if (a1 != Py.None){
                    try{
                        PyObject inter = a1.__int__();
                        if(inter instanceof PyInteger)
                            start = ((PyInteger)inter).getValue();
                        else if(inter instanceof PyLong)
                            start = ((PyLong)inter).getValue().intValue();
                        else
                            start = -1;
                    } catch(PyException x){
                        start = -1;
                    }
                }
		if (a2 != Py.None) {
                    try{
                        PyObject inter = a2.__int__();
                        if(inter instanceof PyInteger)
                            stop = ((PyInteger)inter).getValue();
                        else if(inter instanceof PyLong)
                            stop = ((PyLong)inter).getValue().intValue();
                        else
                            stop = -1;
                    } catch(PyException x){
                        stop = -1;
                    }
		    if (stop == -1) {
		        throw Py.ValueError("Stop argument for islice() must be a non-negative integer or None.");
		    }
		}
	    }
	    if (start < 0 || stop < -1) {
	        throw Py.ValueError("Indices for islice() must be non-negative integers or None.");
	    }

	    if (a3 != null) {
	        if (a3 != Py.None){
                    try{
                        PyObject inter = a3.__int__();
                        if(inter instanceof PyInteger)
                            step = ((PyInteger)inter).getValue();
                        else if(inter instanceof PyLong)
                            step = ((PyLong)inter).getValue().intValue();
                        else
                            step = -1;
                    } catch(PyException x){}
	        }
	    }

	    if (step < 1) {
                throw Py.ValueError("Step for islice() must be a positive integer or None.");
	    }

            PyObject it = seq.__iter__();
            if(it == null)
                return null;
            return new IsliceIterator(type, it, start, stop, step);
        }

	PyObject it; 
        int next, stop, step, cnt;
        IsliceIterator(PyType type, PyObject it, int start, int stop, int step){
	     super(type);
             this.it = it;
             this.next = start;
             this.stop = stop;
             this.step = step;
             this.cnt = 0;
         }

        public PyObject __iternext__(){
	    PyObject item;
	    while(cnt < next){
		item = it.__iternext__();
                if(item == null)
                    return null;
                cnt++;
	    }
            if(stop != -1 && cnt >= stop)
                return null;
            item = it.__iternext__();
	    if (item == null)
		return null;
            cnt++;
            int oldnext = next;
            next += step;
            if(next < oldnext)
                next = stop;
	    return item;
        }
    }

    public static class IzipIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(IzipIterator.class);
        public static final String exposed_name = "itertools.izip";

        public static void typeSetup(PyObject dict,PyType.Newstyle marker) {
             dict.__setitem__("__new__",new PyNewWrapper(IzipIterator.class,"__new__",-1,-1) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     return izip_new(subtype,args,keywords);
                 }
            });
        }

        public static PyObject izip_new(PyType type, PyObject[] iterables, String[] kwords){
            if(type == IzipIterator.type){
                MacroEmulation.PyArg_NoKeywords("izip()", kwords);
             }

            int tuplesize = iterables.length;
            PyObject[] iterators = new PyObject[tuplesize];
	    for (int i=0; i < tuplesize; ++i) {
		PyObject item = iterables[i];
                PyObject it;
                try{
		    it = item.__iter__();
                } catch(PyException x){
		    if(Py.matchException(x, Py.TypeError)){
                        throw Py.TypeError("izip argument #" + (i+1) + " must support iteration");
                    }
                    throw x;
                }
                iterators[i] = it;
	    }

	    return new IzipIterator(type, iterators);
        }

        PyObject[] iterators;
        IzipIterator(PyType type, PyObject[] iterators){
            super(type);
            this.iterators = iterators;
        }

	public PyObject __iternext__(){
           if (iterators.length == 0)
		return null;
           PyObject[] data = new PyObject[iterators.length];
	   for (int i=0 ; i < iterators.length ; i++) {
	       PyObject it = iterators[i];
               PyObject item = it.__iternext__();
	       if (item == null) {
                   return null;
	       }
	       data[i]= item;
	   }
	   return new PyTuple(data);
        }
    }

    public static class RepeatIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(RepeatIterator.class);
        public static final String exposed_name = "itertools.repeat";

        public static void typeSetup(PyObject dict,PyType.Newstyle marker) {
             dict.__setitem__("__new__",new PyNewWrapper(IzipIterator.class,"__new__",1, 2) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     if(!DefaultInfo.check(args.length, info.getMinargs(), info.getMaxargs())){
                         throw info.unexpectedCall(args.length, false);
		     }
                     return repeat_new(subtype,args,keywords);
                 }
            });
        }

        public static PyObject repeat_new(PyType type, PyObject[] args, String[] kwords){
            if(type == RepeatIterator.type){
                MacroEmulation.PyArg_NoKeywords("repeat()", kwords);
            }

	    PyObject element = args[0];
            PyObject cnt = new PyInteger(-1);
            if(args.length == 2){
                PyObject i = args[1];
                boolean throwTypeError = false;
                try{
                    cnt = i.__int__();
                    if(!(cnt instanceof PyInteger) && !(cnt instanceof PyLong))
                        throwTypeError = true;
                } catch(PyException x){
                    throwTypeError = true;
                }
                if(throwTypeError)
                    throw Py.TypeError("an integer is required");
	    }

	    if (args.length == 2 && cnt._lt(Py.Zero).__nonzero__())
	       cnt = Py.Zero;

	    return new RepeatIterator(type, element, cnt);
        }

        PyObject element;
        PyObject cnt;
        RepeatIterator(PyType type, PyObject element, PyObject cnt){
            super(type);
            this.element = element;
            this.cnt = cnt;
        }

	public PyObject __iternext__(){
	    if (cnt._eq(Py.Zero).__nonzero__())
		return null;
            if(cnt._gt(Py.Zero).__nonzero__())
                cnt = cnt._sub(Py.One);
            return element;
        }

        public int __len__(){
            if(cnt._eq(new PyInteger(-1)).__nonzero__()) {
                throw Py.TypeError("len() of unsized object");
	    }
            //This breaks with PyLongs...
	    return ((PyInteger)cnt).getValue();
        }

        public PyString __repr__(){
	    PyString objrepr = element.__repr__();
	    if(objrepr == null) return null;
	    if(cnt._eq(new PyInteger(-1)).__nonzero__()){
		return new PyString("repeat(" + objrepr +")");
            }
	    else{
		return new PyString("repeat("+objrepr+", "+ cnt +")");
            }
        }
    }

    public static class CycleIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(CycleIterator.class);
        public static final String exposed_name = "itertools.cycle";

        public static void typeSetup(PyObject dict,PyType.Newstyle marker) {
             dict.__setitem__("__new__",new PyNewWrapper(CycleIterator.class,"__new__",1, 1) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     if(!DefaultInfo.check(args.length, info.getMinargs(), info.getMaxargs())){
                         throw info.unexpectedCall(args.length, false);
		     }
                     return cycle_new(subtype,args,keywords);
                 }
            });
        }

        public static PyObject cycle_new(PyType type, PyObject[] args, String[] kwords){
            if(type == CycleIterator.type){
                MacroEmulation.PyArg_NoKeywords("cycle()", kwords);
            }
            
            if(args.length == 0) return null;
            PyObject iterable = args[0];

	    /* Get iterator. */
	    PyObject it = iterable.__iter__();
	    if (it == null)
		return null;

	    PyList saved = new PyList();
            return new CycleIterator(type, it, saved);

        }

        PyObject it;
        PyList saved;
        boolean firstpass;
        CycleIterator(PyType type, PyObject it, PyList saved){
            super(type);
            this.it = it;
            this.saved = saved;
            this.firstpass = true;
        }

	public PyObject __iternext__(){
	    PyObject item = null;
	    while (true) {
		item = PyFunctions.PyIter_Next(it);
                if(item != null){
                    if(firstpass)
                        saved.append(item);
                    return item;                    
		}
                if(saved.__len__() == 0) return null;
		it = saved.__iter__();
                if(it == null)
                    return null;
		firstpass = false;
	    }
        }
    }

    public static class GroupbyIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(GroupbyIterator.class);
        public static final String exposed_name = "itertools.groupby";

        public static void typeSetup(PyObject dict,PyType.Newstyle marker) {
             dict.__setitem__("__new__",new PyNewWrapper(GroupbyIterator.class,"__new__",1, 2) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     if(!DefaultInfo.check(args.length, info.getMinargs(), info.getMaxargs())){
                         throw info.unexpectedCall(args.length, false);
		     }
                     return groupby_new(subtype,args,keywords);
                 }
            });
        }

        public static PyObject groupby_new(PyType type, PyObject[] args, String[] kwords){

            String[] paramnames = new String[]{ "iterable", "key" };
            ArgParser ap = new ArgParser("groupby", args, kwords, paramnames, 1);
            PyObject it = ap.getPyObject(0);
            it = it.__iter__();
            if(it == null)
                return null;
            PyObject keyfunc = ap.getPyObject(1, Py.None);
	    return new GroupbyIterator(type, it, keyfunc);

        }

        PyObject it;
        PyObject keyfunc;
        PyObject tgtkey;
        PyObject currkey;
        PyObject currvalue;
        GroupbyIterator(PyType type, PyObject it, PyObject keyfunc){
            super(type);
            this.it = it;
            this.keyfunc = keyfunc;
        }

	public PyObject __iternext__(){
	    PyObject newvalue, newkey, r, grouper;

	    /* skip to next iteration group */
	    for (;;) {
		if (currkey == null);
		else if (tgtkey == null) break;
		else {
                    int rcmp = PyFunctions.PyObject_RichCompareBool(tgtkey, currkey, PyFunctions.EQ);
		    if (rcmp == -1){
                        PyFunctions.ThrowErrorIfSet();
		        return null;
		    }
		    else if (rcmp == 0){
		        break;
                    }
		}
                newvalue = PyFunctions.PyIter_Next(it);
		if (newvalue == null)
		    return null;

		if(keyfunc == Py.None) {
		    newkey = newvalue;
		} else {

                    newkey = keyfunc.__call__(newvalue);
		    if (newkey == null) {
		        return null;
		    }
		}

		currkey = newkey;
		currvalue = newvalue;
	    }

	    tgtkey = currkey;
	    grouper = _grouper_create(this, tgtkey);
	    if (grouper == null) return null;
            return new PyTuple(new PyObject[]{ currkey, grouper});
        } 
    }

    static Grouper _grouper_create(GroupbyIterator parent, PyObject tgtkey){
        return new Grouper(parent, tgtkey);
    }

    static class Grouper extends PyIterator{
	static PyType type = PyType.fromClass(Grouper.class);
        
        GroupbyIterator parent;
        PyObject tgtkey;
        Grouper(GroupbyIterator parent, PyObject tgtkey){
            super(type);
            this.parent = parent;
            this.tgtkey = tgtkey;
	}

        public PyObject __iternext__(){
            GroupbyIterator gbo = parent;
	    PyObject newvalue, newkey, r;
	    int rcmp;
	    if (gbo.currvalue == null) {
                newvalue = PyFunctions.PyIter_Next(gbo.it);
	        if (newvalue == null)
	             return null;

	        if (gbo.keyfunc == Py.None) {
	            newkey = newvalue;
	        } else {
                    newkey = gbo.keyfunc.__call__(newvalue);
                    if(newkey == null)
                        return null;
                }
                Py.assert_(gbo.currkey == null? Py.True: Py.False);
		gbo.currkey = newkey;
		gbo.currvalue = newvalue;
	    }

	    Py.assert_(gbo.currkey != null? Py.True: Py.False);           
            rcmp = PyFunctions.PyObject_RichCompareBool(gbo.tgtkey, gbo.currkey, PyFunctions.EQ);
	    if (rcmp <= 0)
		/* got any error or current group is end */
		return null;

	    r = gbo.currvalue;
	    gbo.currvalue = null;
            gbo.currkey = null;
	    return r;
	}
    }

    public static class DropwhileIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(DropwhileIterator.class);
        public static final String exposed_name = "itertools.dropwhile";

        public static void typeSetup(PyObject dict,PyType.Newstyle marker) {
             dict.__setitem__("__new__",new PyNewWrapper(DropwhileIterator.class,"__new__",2, 2) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     if(!DefaultInfo.check(args.length, info.getMinargs(), info.getMaxargs())){
                         throw info.unexpectedCall(args.length, false);
		     }
                     return dropwhile_new(subtype,args,keywords);
                 }
            });
        }

        public static PyObject dropwhile_new(PyType type, PyObject[] args, String[] kwords){
            if(type == DropwhileIterator.type){
                MacroEmulation.PyArg_NoKeywords("dropwhile()", kwords);
            }

            PyObject func = args[0];
	    PyObject seq = args[1];

	    /* Get iterator. */
	    PyObject it = seq.__iter__();
	    if (it == null)
		return null;
            return new DropwhileIterator(type, func, it);
        }

        PyObject it;
        PyObject func;
        int start;
        DropwhileIterator(PyType type, PyObject func, PyObject it){
            super(type);
            this.func = func;
            this.it = it;
            this.start = 0;
        }

	public PyObject __iternext__(){
	    for (;;) {
		PyObject item = it.__iternext__();
		if (item == null)
		    return null;
		if (start == 1)
		    return item;

                PyObject good = func.__call__(item);
		if (good == null) {
		    return null;
		}
		boolean ok = good.__nonzero__();
		if (!ok) {
		    start = 1;
		    return item;
		}
	    }
        }
    }

    public static class StarmapIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(StarmapIterator.class);
        public static final String exposed_name = "itertools.starmap";

        public static void typeSetup(PyObject dict,PyType.Newstyle marker) {
             dict.__setitem__("__new__",new PyNewWrapper(StarmapIterator.class,"__new__",2, 2) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     if(!DefaultInfo.check(args.length, info.getMinargs(), info.getMaxargs())){
                         throw info.unexpectedCall(args.length, false);
		     }
                     return starmap_new(subtype,args,keywords);
                 }
            });
        }

        public static PyObject starmap_new(PyType type, PyObject[] args, String[] kwords){
            if(type == StarmapIterator.type){
                MacroEmulation.PyArg_NoKeywords("starmap()", kwords);
            }

            PyObject func = args[0];
	    PyObject seq = args[1];

	    /* Get iterator. */
	    PyObject it = seq.__iter__();
	    if (it == null)
		return null;
            return new StarmapIterator(type, func, it);
        }

        PyObject it;
        PyObject func;
        int start;
        StarmapIterator(PyType type, PyObject func, PyObject it){
            super(type);
            this.func = func;
            this.it = it;
            this.start = 0;
        }

	public PyObject __iternext__(){
	    PyObject args = it.__iternext__();
	    if (args == null)
		return null;
	    if (!(args instanceof PyTuple)) {
	        throw Py.TypeError("iterator must return a tuple");
	    }
	    PyObject result = func.__call__(((PyTuple)args).getArray());
	    return result;
        }
    }


    private static final int LINKCELLS = 57;

    public static class TeeIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(TeeIterator.class);
        public static final String exposed_name = "itertools.tee";

        public static void typeSetup(PyObject dict,PyType.Newstyle marker){
            try{
	        PyBuiltinFunction function =  Boilerplate.newNarrow(TeeIterator.class, "tee_copy", new int[]{0}, new boolean[]{ false }, null);
		dict.__setitem__("__copy__", new PyMethodDescr("__cal__", TeeIterator.class, 0, 0, function));
            } catch(Exception x){}
            dict.__setitem__("__new__",new PyNewWrapper(TeeIterator.class,"__new__",1, 1) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     if(!DefaultInfo.check(args.length, info.getMinargs(), info.getMaxargs())){
                         throw info.unexpectedCall(args.length, false);
		     }
                     return tee_new(subtype,args,keywords);
                 }
            });
        }

        public static PyObject tee_new(PyType type, PyObject[] iterable, String[] kwords){
            //if(type == TeeIterator.type){
	    //  MacroEmulation.PyArg_NoKeywords("tee()", kwords);
            //}
            return tee_fromiterable(iterable[0]);
        }

        static PyObject tee_fromiterable(PyObject iterable){

            PyObject to;
	    PyObject it = iterable.__iter__();
	    if (it == null)
	        return null;

            if(it.getType().isSubType(type)){
                return ((TeeIterator)it).tee_copy();
	    }

            TeeDataObject dataobj = TeeDataObject.teedataobject_new(it);
            if(dataobj == null)
                return null;

            return new TeeIterator(type, dataobj);
        }


        TeeDataObject dataobj;
        int index;
        TeeIterator(PyType type, TeeDataObject dataobj){
            super(type);
            this.dataobj = dataobj;
            this.index = 0;
        }

	public PyObject __iternext__(){
	    if (index >= LINKCELLS) {
		TeeDataObject link = dataobj.teedataobject_jumplink();
		dataobj = link;
		index = 0;
	    }
	    PyObject value = dataobj.teedataobject_getitem(index);
	    if (value == null)
	        return null;
	    index++;
	    return value;
        }

        public PyObject tee_copy(){
            TeeIterator newto = new TeeIterator(type, dataobj);
            newto.index = index;
  	    return newto;
        }
    }

    static class TeeDataObject extends PyObject{
	public static PyType type = PyType.fromClass(TeeDataObject.class);
        public static String exposed_name = "itertools.tee_dataobject";

        static TeeDataObject teedataobject_new(PyObject iterable){
            return new TeeDataObject(iterable);
        }

        PyObject it;
        int numread;
        TeeDataObject nextlink;
        PyObject[] values;
        TeeDataObject(PyObject it){
            super(type);
            this.it = it;
            numread = 0;
            nextlink = null;
            values = new PyObject[LINKCELLS];
        }

        PyObject teedataobject_getitem(int i){
	    PyObject value;
	    Py.assert_(i < LINKCELLS? Py.True: Py.False);
	    if (i < numread)
		value = values[i];
	    else {
		/* this is the lead iterator, so fetch more data */
		Py.assert_(i == numread? Py.True: Py.False);
		value = PyFunctions.PyIter_Next(it);
		if (value == null)
		    return null;
		numread++;
		values[i] = value;
	    }
	    return value;
        }

        public TeeDataObject teedataobject_jumplink() {
	    if(nextlink == null)
		nextlink = teedataobject_new(it);
	    return nextlink;
        }


    }

    static class Tee extends PyBuiltinFunctionWide{

        PyObject self;
        Tee(PyObject self, Info info){
            super(info);
            this.self = self;
	}

        public PyBuiltinFunction makeBound(PyObject self){
	    return new Tee(self, info);
        }

	public PyObject inst_call(PyObject self, PyObject[] args){
	    //Py_ssize_t i, n=2;
            if(!DefaultInfo.check(args.length, info.getMinargs(), info.getMaxargs())){
                throw info.unexpectedCall(args.length, false);
	    }            
            PyObject n = new PyInteger(2);
	    PyObject it, iterable, copyable, result;
            if(args.length == 0)
                return null;
            iterable = args[0];
            if(args.length > 1){
                boolean throwTypeError = false;
                try{
                    n = args[1].__int__();
                    if(!(n instanceof PyInteger) && !(n instanceof PyLong))
                        throwTypeError = true;
                    } catch(PyException x){
                        throwTypeError = true;
                }
                if(throwTypeError)
                    throw Py.TypeError("an integer is required");
            }

	    if (n._lt(Py.Zero).__nonzero__()) {
		throw Py.ValueError("n must be >= 0");
	    }

	    if (n._eq(Py.Zero).__nonzero__())
	        return new PyTuple();
	    it = iterable.__iter__(); 
	    if (it == null) {
		return null;
	    }

            if(!__builtin__.hasattr(it, new PyString("__copy__"))){
		copyable = TeeIterator.tee_fromiterable(it);
		if (copyable == null) {
		    return null;
		}
	    } else
		copyable = it;
           
	    int size = n instanceof PyInteger? ((PyInteger)n).getValue(): ((PyLong)n).getValue().intValue();
            PyObject[] result_args = new PyObject[size];
            result_args[0] = copyable;
	    for (int i=1 ; i< size ; i++) {
                copyable = copyable.invoke("__copy__");
		if (copyable == null) {
		    return null;
		}
                result_args[i] = copyable;
	   }
	   return new PyTuple(result_args);
        }
    }

    public static class TakewhileIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(TakewhileIterator.class);
        public static final String exposed_name = "itertools.takewhile";

        public static void typeSetup(PyObject dict,PyType.Newstyle marker) {
             dict.__setitem__("__new__",new PyNewWrapper(TakewhileIterator.class,"__new__",2,2) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     if(!DefaultInfo.check(args.length, info.getMinargs(), info.getMaxargs())){
                         throw info.unexpectedCall(args.length, false);
		     }
                     return takewhile_new(subtype,args,keywords);
                 }
            });
        }

        public static PyObject takewhile_new(PyType type, PyObject[] args, String[] kwords){
            if(type == TakewhileIterator.type){
                MacroEmulation.PyArg_NoKeywords("takewhile()", kwords);
             }

	    PyObject func = args[0];
            PyObject seq = args[1];


	    PyObject it = seq.__iter__();
	    if (it == null)
		return null;

            return new TakewhileIterator(type, func, it);
        }

        PyObject it;
        PyObject func;
        int stop;
        TakewhileIterator(PyType type, PyObject func, PyObject it){
            super(type);
            this.func = func;
            this.it = it;
            this.stop = 0;
        }

	public PyObject __iternext__(){
	    if(stop == 1)
		return null;

	    PyObject item = it.__iternext__();
	    if (item == null)
		return null;

	    PyObject good = func.__call__(item);
	    if (good == null) {
		return null;
	    }
	    boolean ok = good.__nonzero__();
	    if (ok)
	        return item;
	    stop = 1;
	    return null;
        }
    }

    public static class IfilterIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(IfilterIterator.class);
        public static final String exposed_name = "itertools.ifilter";

        public static void typeSetup(PyObject dict,PyType.Newstyle marker) {
             dict.__setitem__("__new__",new PyNewWrapper(IfilterIterator.class,"__new__",2,2) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     if(!DefaultInfo.check(args.length, info.getMinargs(), info.getMaxargs())){
                         throw info.unexpectedCall(args.length, false);
		     }
                     return ifilter_new(subtype,args,keywords);
                 }
            });
        }

        public static PyObject ifilter_new(PyType type, PyObject[] args, String[] kwords){
            if(type == IfilterIterator.type){
                MacroEmulation.PyArg_NoKeywords("ifilter()", kwords);
	    }

            PyObject func = args[0];
            PyObject seq = args[1];

	    /* Get iterator. */
	    PyObject it = seq.__iter__();
	    if (it == null)
	        return null;
            return new IfilterIterator(type, func, it);
        }

        PyObject it;
        PyObject func;
        IfilterIterator(PyType type, PyObject func, PyObject it){
            super(type);
            this.func = func;
            this.it = it;
        }

	public PyObject __iternext__(){
	    PyObject item;
	    boolean ok;
	    for (;;) {
		item = it.__iternext__();
		if (item == null)
		    return null;

		if (func == Py.None) {
		    ok = item.__nonzero__();
		} else {
		    PyObject good = func.__call__(item);
		    if (good == null) {
		        return null;
		    }
		    ok = good.__nonzero__();
		}
		if (ok)
		    return item;
            }	   
        }
    }

    public static class IfilterfalseIterator extends IterToolsBaseIterator{
        static PyType type = PyType.fromClass(IfilterfalseIterator.class);
        public static final String exposed_name = "itertools.ifilterfalse";

        public static void typeSetup(PyObject dict,PyType.Newstyle marker) {
             dict.__setitem__("__new__",new PyNewWrapper(IfilterfalseIterator.class,"__new__",2,2) {
                 public PyObject new_impl(boolean init,PyType subtype,PyObject[]args,String[]keywords) {
                     if(!DefaultInfo.check(args.length, info.getMinargs(), info.getMaxargs())){
                         throw info.unexpectedCall(args.length, false);
		     }
                     return ifilterfalse_new(subtype,args,keywords);
                 }
            });
        }

        public static PyObject ifilterfalse_new(PyType type, PyObject[] args, String[] kwords){
            if(type == IfilterIterator.type){
                MacroEmulation.PyArg_NoKeywords("ifilterfalse()", kwords);
	    }

            PyObject func = args[0];
            PyObject seq = args[1];

	    /* Get iterator. */
	    PyObject it = seq.__iter__();
	    if (it == null)
	        return null;
            return new IfilterfalseIterator(type, func, it);
        }

        PyObject it;
        PyObject func;
        IfilterfalseIterator(PyType type, PyObject func, PyObject it){
            super(type);
            this.func = func;
            this.it = it;
        }

	public PyObject __iternext__(){
	    PyObject item;
	    boolean ok;
	    for (;;) {
		item = it.__iternext__();
		if (item == null)
		    return null;

		if (func == Py.None) {
		    ok = item.__nonzero__();
		} else {
		    PyObject good = func.__call__(item);
		    if (good == null) {
		        return null;
		    }
		    ok = good.__nonzero__();
		}
		if (!ok)
		    return item;
            }	   
        }
    }


    public static abstract class IterToolsBaseIterator extends PyIterator{
        public IterToolsBaseIterator(PyType type){
            super(type);
        }
        public void dispatch__init__(PyType type,PyObject[]args,String[]keywords) {
            PyType self_type=getType();
            if (self_type.isSubType(type)) {
                PyObject impl=self_type.lookup("__init__");
                if (impl!=null)
                    impl.__get__(this,self_type).__call__(args,keywords);
            }
        }
    }
}