package org.jythonx.utils;

import org.python.core.PyException;
import org.python.core.PyInteger;
import org.python.core.PyLong;
import org.python.core.PyObject;
import org.python.core.Py;
import org.python.core.PySystemState;
import org.python.core.PyType;
import org.python.core.__builtin__;

public final class CoreUtils{

    static PyInteger maxint = new PyInteger(PySystemState.maxint);
    static PyInteger minint = new PyInteger(PySystemState.minint);
    public static final PyObject[] sliceStartStopMunge(PyObject sliced, PyObject start, PyObject stop){
        if(start._lt(Py.Zero).__nonzero__()){
            start = __builtin__.max(new PyObject[]{start, minint}, Py.NoKeywords);
            int len = sliced.__len__();
            start = start.__add__(new PyInteger(len));
	}
        if(stop._lt(Py.Zero).__nonzero__()){
            stop = __builtin__.max(new PyObject[]{stop, minint}, Py.NoKeywords);
            int len = sliced.__len__();                
            stop = stop.__add__(new PyInteger(len));
	}
        start = __builtin__.min(new PyObject[]{start, maxint}, Py.NoKeywords);
        stop = __builtin__.min(new PyObject[]{stop, maxint}, Py.NoKeywords);
        return new PyObject[]{ start, stop };
    }

    /** Fix the overflow */
    public static int safeIndexConvertToInteger(PyObject obj, String errormessage){
        PyObject index = obj.__index__();
        try {
	    return ((PyInteger)index.__int__()).getValue();
        } catch (PyException exc) {
            if(obj instanceof PyLong){
                if (Py.matchException(exc, Py.OverflowError)) {
                    if (new PyLong(0L).__cmp__(obj) < 0){
                        return Integer.MAX_VALUE;
                    }
                    else{
                        return 0;
                    }
                }
            }
        }
        throw Py.TypeError(errormessage);
    }

    /** Let the overflow occur */
    public static int unsafeIndexConvertToInteger(PyObject obj){
	PyObject index = obj.__index__();
        return ((PyInteger)index.__int__()).getValue();
    }

    public static PyObject __getslice__(PyObject impl, PyObject this_, PyType self_type, PyObject start, PyObject stop){
        try {
            PyObject[] startstop = sliceStartStopMunge(this_, start, stop);
            return impl.__get__(this_,self_type).__call__(startstop[0],startstop[1]);
        } catch (PyException exc) {
            if (Py.matchException(exc,Py.LookupError))
                return null;
            throw exc;
        }
    }


}