package org.jythonx.utils;

import org.python.core.*;

public final class PyFunctions{
    private PyFunctions(){}

    public static final int EQ = 0, LT = 1, LE=2, NE=3, GT=4, GE=5;

    public static PyObject PyIter_Next(PyObject iter){
        try{
            return iter.__iternext__();
        } catch(PyException x){
            if(Py.matchException(x, Py.StopIteration)){
                PyErr_Clear();
                return null;
            }
	    throw x;
        }
    }

    public static int PyObject_RichCompareBool(PyObject one, PyObject two, int op){
        int rcmp = 0;
        try{
            PyObject rv;
            switch(op){
	        case EQ:
                    rv = one._eq(two);
                    break;
	        case LT:
                    rv = one._lt(two);
                    break;
	        case LE:
                    rv = one._le(two);
                    break;
	        case NE:
                    rv = one._ne(two);
                    break;
	        case GT:
                    rv = one._gt(two);
                    break;
	        case GE:
                    rv = one._gt(two);
                    break;
	        default:
		    throw new IllegalArgumentException("op not valid");
	    }
            if(rv == null) rcmp = -1;
            else if(rv.__nonzero__())
                rcmp = 1;                        
        } catch(PyException x){
            Py.setException(x, Py.getFrame());
            return -1;
        }
        return rcmp;
    }

    public static int PyObject_Hash(PyObject obj){
        try{
            return obj.__hash__().getValue();
	} catch(PyException x){
            Py.setException(x, Py.getFrame());
            return -1; 
        }
    }

    public static boolean PyErr_ExceptionMatches(PyObject x){
	return Py.matchException(Py.getThreadState().exception,x );
    }

    public static boolean PyErr_Occurred(){
        return Py.getThreadState().exception != null;
    }

    public static void PyErr_Clear(){
	Py.getThreadState().exception = null;
    }

    public static void ThrowErrorIfSet(){
        PyException exc = Py.getThreadState().exception;
        if(exc != null)
            throw exc;
    }


}